# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import uuid

from oslo_log import log as logging
from six.moves import range

from congress.datalog import compile

LOG = logging.getLogger(__name__)


# A unifier designed for the bi_unify_atoms routine
# which is used by a backward-chaining style datalog implementation.
# Main goal: minimize memory allocation by manipulating only unifiers
#   to keep variable namespaces separate.

class BiUnifier(object):
    """A unifier designed for bi_unify_atoms.

    Recursive datastructure. When adding a binding variable u to
    variable v, keeps a reference to the unifier for v.
    A variable's identity is its name plus its unification context.
    This enables a variable with the same name but from two
    different atoms to be treated as different variables.
    """
    class Value(object):
        def __init__(self, value, unifier):
            # actual value
            self.value = value
            # unifier context
            self.unifier = unifier

        def __str__(self):
            return "<{},{}>".format(
                str(self.value), repr(self.unifier))

        def recur_str(self):
            if self.unifier is None:
                recur = str(self.unifier)
            else:
                recur = self.unifier.recur_str()
            return "<{},{}>".format(
                str(self.value), recur)

        def __eq__(self, other):
            return self.value == other.value and self.unifer == other.unifier

        def __repr__(self):
            return "Value(value={}, unifier={})".format(
                repr(self.value), repr(self.unifier))

    class Undo(object):
        def __init__(self, var, unifier):
            self.var = var
            self.unifier = unifier

        def __str__(self):
            return "<var: {}, unifier: {}>".format(
                str(self.var), str(self.unifier))

        def __eq__(self, other):
            return self.var == other.var and self.unifier == other.unifier

    def __init__(self, dictionary=None):
        # each value is a Value
        self.contents = {}
        if dictionary is not None:
            for var, value in dictionary.items():
                self.add(var, value, None)

    def add(self, var, value, unifier):
        value = self.Value(value, unifier)
        # LOG.debug("Adding %s -> %s to unifier %s", var, value, self)
        self.contents[var] = value
        return self.Undo(var, self)

    def delete(self, var):
        if var in self.contents:
            del self.contents[var]

    def value(self, term):
        if term in self.contents:
            return self.contents[term]
        else:
            return None

    def apply(self, term, caller=None):
        return self.apply_full(term, caller=caller)[0]

    def apply_full(self, term, caller=None):
        """Recursively apply unifiers to TERM.

        Return (i) the final value and (ii) the final unifier.
        If the final value is a variable, instantiate
        with a new variable if not in KEEP_VARS
        """
        # LOG.debug("apply_full(%s, %s)", term, self)
        val = self.value(term)
        if val is None:
            # If result is a variable and this variable is not one of those
            #   in the top-most calling context, then create a new variable
            #   name based on this Binding.
            # This process avoids improper variable capture.
            #   Outputting the same variable with the same binding twice will
            #   generate the same output, but outputting the same variable with
            #   different bindings will generate different outputs.
            # Note that this variable name mangling
            #   is not done for the top-most variables,
            #   which makes output a bit easier to read.
            # Unfortunately, the process is non-deterministic from one run
            #   to the next, which makes testing difficult.
            if (caller is not None and term.is_variable() and
                    not (term in caller.variables and caller.binding is self)):
                return (compile.Variable(term.name + str(id(self))), self)
            else:
                return (term, self)
        elif val.unifier is None or not val.value.is_variable():
            return (val.value, val.unifier)
        else:
            return val.unifier.apply_full(val.value)

    def is_one_to_one(self):
        image = set()  # set of all things mapped TO
        for x in self.contents:
            val = self.apply(x)
            if val in image:
                return False
            image.add(val)
        return True

    def __str__(self):
        s = repr(self)
        s += "={"
        s += ",".join(["{}:{}".format(str(var), str(val))
                       for var, val in self.contents.items()])
        s += "}"
        return s

    def recur_str(self):
        s = repr(self)
        s += "={"
        s += ",".join(["{}:{}".format(var, val.recur_str())
                       for var, val in self.contents.items()])
        s += "}"
        return s

    def __eq__(self, other):
        return self.contents == other.contents


def binding_str(binding):
    """Handles string conversion of either dictionary or Unifier."""
    if isinstance(binding, dict):
        s = ",".join(["{}: {}".format(str(var), str(val))
                     for var, val in binding.items()])
        return '{' + s + '}'
    else:
        return str(binding)


def undo_all(changes):
    """Undo all the changes in CHANGES."""
    # LOG.debug("undo_all(%s)",
    #     "[" + ",".join([str(x) for x in changes]) + "]")
    if changes is None:
        return
    for change in changes:
        if change.unifier is not None:
            change.unifier.delete(change.var)


def same_schema(atom1, atom2, theoryname=None):
    """Return True if ATOM1 and ATOM2 have the same schema.

    THEORYNAME is the default theory name.
    """
    if not atom1.table.same(atom2.table, theoryname):
        return False
    if len(atom1.arguments) != len(atom2.arguments):
        return False
    return True


def bi_unify_atoms(atom1, unifier1, atom2, unifier2, theoryname=None):
    """Unify atoms.

    If possible, modify BiUnifier UNIFIER1 and BiUnifier UNIFIER2 so that
    ATOM1.plug(UNIFIER1) == ATOM2.plug(UNIFIER2).
    Returns None if not possible; otherwise, returns
    a list of changes to unifiers that can be undone
    with undo-all. May alter unifiers besides UNIFIER1 and UNIFIER2.
    THEORYNAME is the default theory name.
    """
    # logging.debug("Unifying %s under %s and %s under %s",
    #      atom1, unifier1, atom2, unifier2)
    if not same_schema(atom1, atom2, theoryname):
        return None
    return bi_unify_lists(atom1.arguments, unifier1,
                          atom2.arguments, unifier2)


def bi_unify_lists(iter1, unifier1, iter2, unifier2):
    """Unify lists.

    If possible, modify BiUnifier UNIFIER1 and BiUnifier UNIFIER2 such that
    iter1.plug(UNIFIER1) == iter2.plug(UNIFIER2), assuming PLUG is defined
    over lists.  Returns None if not possible; otherwise, returns
    a list of changes to unifiers that can be undone
    with undo-all. May alter unifiers besides UNIFIER1 and UNIFIER2.
    """
    if len(iter1) != len(iter2):
        return None
    changes = []
    for i in range(0, len(iter1)):
        assert isinstance(iter1[i], compile.Term)
        assert isinstance(iter2[i], compile.Term)
        # grab values for args
        val1, binding1 = unifier1.apply_full(iter1[i])
        val2, binding2 = unifier2.apply_full(iter2[i])
        # logging.debug("val(%s)=%s at %s, val(%s)=%s at %s",
        #     atom1.arguments[i], val1, binding1,
        #     atom2.arguments[i], val2, binding2)
        # assign variable (if necessary) or fail
        if val1.is_variable() and val2.is_variable():
            # logging.debug("1 and 2 are variables")
            if bi_var_equal(val1, binding1, val2, binding2):
                continue
            else:
                changes.append(binding1.add(val1, val2, binding2))
        elif val1.is_variable() and not val2.is_variable():
            # logging.debug("Left arg is a variable")
            changes.append(binding1.add(val1, val2, binding2))
        elif not val1.is_variable() and val2.is_variable():
            # logging.debug("Right arg is a variable")
            changes.append(binding2.add(val2, val1, binding1))
        elif val1 == val2:
            continue
        else:
            # logging.debug("Unify failure: undoing")
            undo_all(changes)
            return None
    return changes

# def plug(atom, binding, withtable=False):
#     """ Returns a tuple representing the arguments to ATOM after having
#         applied BINDING to the variables in ATOM. """
#     if withtable is True:
#         result = [atom.table]
#     else:
#         result = []
#     for i in range(0, len(atom.arguments)):
#         if (atom.arguments[i].is_variable() and
#             atom.arguments[i].name in binding):
#             result.append(binding[atom.arguments[i].name])
#         else:
#             result.append(atom.arguments[i].name)
#     return tuple(result)


def match_tuple_atom(tupl, atom):
    """Get bindings.

    Returns a binding dictionary that when applied to ATOM's arguments
    gives exactly TUPLE, or returns None if no such binding exists.
    """
    if len(tupl) != len(atom.arguments):
        return None
    binding = {}
    for i in range(0, len(tupl)):
        arg = atom.arguments[i]
        if arg.is_variable():
            if arg.name in binding:
                oldval = binding[arg.name]
                if oldval != tupl[i]:
                    return None
            else:
                binding[arg.name] = tuple[i]
    return binding


def match_atoms(atom1, unifier, atom2):
    """Modify UNIFIER so that ATOM1.plug(UNIFIER) == ATOM2.

    ATOM2 is assumed to be ground.
    UNIFIER is assumed to be a BiUnifier.
    Return the changes to UNIFIER or None if matching is impossible.

    Matching is a special case of instance-checking since ATOM2
    in this case must be ground, whereas there is no such limitation
    for instance-checking.  This makes the code significantly simpler
    and faster.
    """
    if not same_schema(atom1, atom2):
        return None
    changes = []
    for i in range(0, len(atom1.arguments)):
        val, binding = unifier.apply_full(atom1.arguments[i])
        # LOG.debug("val(%s)=%s at %s; comparing to object %s",
        #     atom1.arguments[i], val, binding, atom2.arguments[i])
        if val.is_variable():
            changes.append(binding.add(val, atom2.arguments[i], None))
        else:
            if val.name != atom2.arguments[i].name:
                undo_all(changes)
                return None
    return changes


def bi_var_equal(var1, unifier1, var2, unifier2):
    """Check var equality.

    Returns True iff variable VAR1 in unifier UNIFIER1 is the same
    variable as VAR2 in UNIFIER2.
    """
    return (var1 == var2 and unifier1 is unifier2)


def same(formula1, formula2):
    """Check formulas are the same.

    Determine if FORMULA1 and FORMULA2 are the same up to a variable
    renaming. Treats FORMULA1 and FORMULA2 as having different
    variable namespaces. Returns None or the pair of unifiers.
    """
    if isinstance(formula1, compile.Literal):
        if isinstance(formula2, compile.Rule):
            return None
        elif formula1.is_negated() != formula2.is_negated():
            return None
        else:
            u1 = BiUnifier()
            u2 = BiUnifier()
            if same_atoms(formula1, u1, formula2, u2, set()) is not None:
                return (u1, u2)
            return None
    elif isinstance(formula1, compile.Rule):
        if isinstance(formula2, compile.Literal):
            return None
        else:
            if len(formula1.body) != len(formula2.body):
                return None
            u1 = BiUnifier()
            u2 = BiUnifier()
            bound2 = set()
            result = same_atoms(formula1.head, u1, formula2.head, u2, bound2)
            if result is None:
                return None
            for i in range(0, len(formula1.body)):
                result = same_atoms(
                    formula1.body[i], u1, formula2.body[i], u2, bound2)
                if result is None:
                    return None
            return (u1, u2)
    else:
        return None


def same_atoms(atom1, unifier1, atom2, unifier2, bound2):
    """Check whether atoms are identical.

    Modifies UNIFIER1 and UNIFIER2 to demonstrate
    that ATOM1 and ATOM2 are identical up to a variable renaming.
    Returns None if not possible or the list of changes if it is.
    BOUND2 is the set of variables already bound in UNIFIER2
    """
    def die():
        undo_all(changes)
        return None
    LOG.debug("same_atoms(%s, %s)", atom1, atom2)
    if not same_schema(atom1, atom2):
        return None
    changes = []
    # LOG.debug("same_atoms entering loop")
    for i in range(0, len(atom1.arguments)):
        val1, binding1 = unifier1.apply_full(atom1.arguments[i])
        val2, binding2 = unifier2.apply_full(atom2.arguments[i])
        # LOG.debug("val1: %s at %s; val2: %s at %s",
        #     val1, binding1, val2, binding2)
        if val1.is_variable() and val2.is_variable():
            if bi_var_equal(val1, binding1, val2, binding2):
                continue
            # if we already bound either of these variables, not SAME
            if not bi_var_equal(val1, binding1, atom1.arguments[i], unifier1):
                # LOG.debug("same_atoms: arg1 already bound")
                return die()
            if not bi_var_equal(val2, binding2, atom2.arguments[i], unifier2):
                # LOG.debug("same_atoms: arg2 already bound")
                return die()
            if val2 in bound2:
                # LOG.debug("same_atoms: binding is not 1-1")
                return die()
            changes.append(binding1.add(val1, val2, binding2))
            bound2.add(val2)
        elif val1.is_variable():
            # LOG.debug("val1 is a variable")
            return die()
        elif val2.is_variable():
            # LOG.debug("val2 is a variable")
            return die()
        elif val1 != val2:
            # one is a variable and one is not or unmatching object constants
            # LOG.debug("val1 != val2")
            return die()
    return changes


def instance(formula1, formula2):
    """Determine if FORMULA1 is an instance of FORMULA2.

    If there is some binding that when applied to FORMULA1 results
    in FORMULA2. Returns None or a unifier.
    """
    LOG.debug("instance(%s, %s)", formula1, formula2)
    if isinstance(formula1, compile.Literal):
        if isinstance(formula2, compile.Rule):
            return None
        elif formula1.is_negated() != formula2.is_negated():
            return None
        else:
            u = BiUnifier()
            if instance_atoms(formula1, formula2, u) is not None:
                return u
            return None
    elif isinstance(formula1, compile.Rule):
        if isinstance(formula2, compile.Literal):
            return None
        else:
            if len(formula1.body) != len(formula2.body):
                return None
            u = BiUnifier()
            result = instance_atoms(formula1.head, formula2.head, u)
            if result is None:
                return None
            for i in range(0, len(formula1.body)):
                result = same_atoms(
                    formula1.body[i], formula2.body[i], u)
                if result is None:
                    return None
            return u
    else:
        return None


def instance_atoms(atom1, atom2, unifier2):
    """Check atoms equality by adding bindings.

    Adds bindings to UNIFIER2 to make ATOM1 equal to ATOM2
    after applying UNIFIER2 to ATOM2 only.   Returns None if
    no such bindings make equality hold.
    """
    def die():
        undo_all(changes)
        return None
    LOG.debug("instance_atoms(%s, %s)", atom1, atom2)
    if not same_schema(atom1, atom2):
        return None
    unifier1 = BiUnifier()
    changes = []
    for i in range(0, len(atom1.arguments)):
        val1, binding1 = unifier1.apply_full(atom1.arguments[i])
        val2, binding2 = unifier2.apply_full(atom2.arguments[i])
        # LOG.debug("val1: %s at %s; val2: %s at %s",
        #     val1, binding1, val2, binding2)
        if val1.is_variable() and val2.is_variable():
            if bi_var_equal(val1, binding1, val2, binding2):
                continue
            # if we already bound either of these variables, not INSTANCE
            if not bi_var_equal(val1, binding1, atom1.arguments[i], unifier1):
                # LOG.debug("instance_atoms: arg1 already bound")
                return die()
            if not bi_var_equal(val2, binding2, atom2.arguments[i], unifier2):
                # LOG.debug("instance_atoms: arg2 already bound")
                return die()
            # add binding to UNIFIER2
            changes.append(binding2.add(val2, val1, binding1))
        elif val1.is_variable():
            return die()
        elif val2.is_variable():
            changes.append(binding2.add(val2, val1, binding1))
            # LOG.debug("var2 is a variable")
        elif val1 != val2:
            # unmatching object constants
            # LOG.debug("val1 != val2")
            return die()
    return changes


def skolemize(formulas):
    """Instantiate all variables consistently with UUIDs in the formulas."""
    # create binding then plug it in.
    variables = set()
    for formula in formulas:
        variables |= formula.variables()

    binding = {}
    for var in variables:
        binding[var] = compile.Term.create_from_python(str(uuid.uuid4()))

    return [formula.plug(binding) for formula in formulas]



# Copyright 2015 Huawei.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import itertools

import congress.common.config
import congress.exception
import congress.utils


def list_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             congress.common.config.core_opts,
             congress.utils.utils_opts,
             congress.exception.exc_log_opts,
         )),
    ]



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import utility


LOG = logging.getLogger(__name__)


class FactSet(object):
    """FactSet

    Maintains a set of facts, and provides indexing for efficient iteration,
    given a partial or full match.  Expects that all facts are the same width.
    """

    def __init__(self):
        self._facts = utility.OrderedSet()

        # key is a sorted tuple of column indices, values are dict mapping a
        # specific value for the key to a set of Facts.
        self._indicies = {}

    def __contains__(self, fact):
        return fact in self._facts

    def __len__(self):
        return len(self._facts)

    def __iter__(self):
        return self._facts.__iter__()

    def add(self, fact):
        """Add a fact to the FactSet

        Returns True if the fact is absent from this FactSet and adds the
        fact, otherwise returns False.
        """
        assert isinstance(fact, tuple)
        changed = self._facts.add(fact)
        if changed:
            # Add the fact to the indicies
            try:
                for index in self._indicies.keys():
                    self._add_fact_to_index(fact, index)
            except Exception as e:
                self._facts.discard(fact)
                raise e
        return changed

    def remove(self, fact):
        """Remove a fact from the FactSet

        Returns True if the fact is in this FactSet and removes the fact,
        otherwise returns False.
        """
        changed = self._facts.discard(fact)
        if changed:
            # Remove from indices
            try:
                for index in self._indicies.keys():
                    self._remove_fact_from_index(fact, index)
            except Exception as e:
                self._facts.add(fact)
                raise e
        return changed

    def create_index(self, columns):
        """Create an index

        @columns is a tuple of column indicies that index into the facts in
        self.  @columns must be sorted in ascending order, and each column
        index must be less than the width of a fact in self.  If the index
        exists, do nothing.
        """
        assert sorted(columns) == list(columns)
        assert len(columns)

        if columns in self._indicies:
            return

        for f in self._facts:
            self._add_fact_to_index(f, columns)

    def remove_index(self, columns):
        """Remove an index

        @columns is a tuple of column indicies that index into the facts in
        self.  @columns must be sorted in ascending order, and each column
        index must be less than the width of a fact in self.  If the index
        does not exists, raise KeyError.
        """
        assert sorted(columns) == list(columns)
        if columns in self._indicies:
            del self._indicies[columns]

    def has_index(self, columns):
        """Returns True if the index exists."""
        return columns in self._indicies

    def find(self, partial_fact, iterations=None):
        """Find Facts given a partial fact

        @partial_fact is a tuple of pair tuples.  The first item in each
        pair tuple is an index into a fact, and the second item is a value to
        match again self._facts.  Expects the pairs to be sorted by index in
        ascending order.

        @iterations is either an empty list or None.  If @iterations is an
        empty list, then find() will append the number of iterations find()
        used to compute the return value(this is useful for testing indexing).

        Returns matching Facts.
        """
        index = tuple([i for i, v in partial_fact])
        k = tuple([v for i, v in partial_fact])
        if index in self._indicies:
            if iterations is not None:
                iterations.append(1)
            if k in self._indicies[index]:
                return self._indicies[index][k]
            else:
                return set()

        # There is no index, so iterate.
        matches = set()
        for f in self._facts:
            match = True
            for i, v in partial_fact:
                if f[i] != v:
                    match = False
                    break
            if match:
                matches.add(f)

        if iterations is not None:
            iterations.append(len(self._facts))
        return matches

    def _compute_key(self, columns, fact):
        # assumes that @columns is sorted in ascending order.
        return tuple([fact[i] for i in columns])

    def _add_fact_to_index(self, fact, index):
        if index not in self._indicies:
            self._indicies[index] = {}

        k = self._compute_key(index, fact)
        if k not in self._indicies[index]:
            self._indicies[index][k] = set((fact,))
        else:
            self._indicies[index][k].add(fact)

    def _remove_fact_from_index(self, fact, index):
        k = self._compute_key(index, fact)
        self._indicies[index][k].remove(fact)
        if not len(self._indicies[index][k]):
            del self._indicies[index][k]



# Copyright (c) 2015 Hewlett-Packard. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import inspect

import keystoneauth1.identity.v2 as ksidentity
import keystoneauth1.session as kssession
import keystoneclient.v2_0.client as ksclient
import muranoclient.client
from muranoclient.common import exceptions as murano_exceptions
from oslo_log import log as logging
from oslo_utils import uuidutils
import six

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils
from congress.datasources import murano_classes
from congress import utils


logger = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return MuranoDriver(name, keys, inbox, datapath, args)


class MuranoDriver(datasource_driver.PollingDataSourceDriver,
                   datasource_driver.ExecutionDriver):
    OBJECTS = "objects"
    PARENT_TYPES = "parent_types"
    PROPERTIES = "properties"
    RELATIONSHIPS = "relationships"
    CONNECTED = "connected"
    STATES = "states"
    ACTIONS = "actions"
    UNUSED_PKG_PROPERTIES = ['id', 'owner_id', 'description']
    UNUSED_ENV_PROPERTIES = ['id', 'tenant_id']
    APPS_TYPE_PREFIXES = ['io.murano.apps', 'io.murano.databases']

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(MuranoDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        logger.debug("Credentials = %s" % self.creds)
        # TODO(ekcs): factor session creation out of individual driver
        #   One single keystone session can be shared by all drivers
        auth = ksidentity.Password(
            auth_url=self.creds['auth_url'],
            username=self.creds['username'],
            password=self.creds['password'],
            tenant_name=self.creds['tenant_name'])
        session = kssession.Session(auth=auth)
        keystone = ksclient.Client(**self.creds)
        murano_endpoint = keystone.service_catalog.url_for(
            service_type='application-catalog',
            endpoint_type='publicURL')
        logger.debug("murano_endpoint = %s" % murano_endpoint)
        client_version = "1"
        self.murano_client = muranoclient.client.Client(
            client_version, murano_endpoint, session=session)
        self.add_executable_client_methods(
            self.murano_client,
            'muranoclient.v1.')
        logger.debug("Successfully created murano_client")

        self.action_call_returns = []
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'murano'
        result['description'] = ('Datasource driver that interfaces with '
                                 'murano')
        result['config'] = datasource_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        """Called when it is time to pull new data from this datasource.

        Sets self.state[tablename] = <set of tuples of strings/numbers>
        for every tablename exported by this datasource.
        """
        self.state[self.STATES] = set()
        self.state[self.OBJECTS] = set()
        self.state[self.PROPERTIES] = set()
        self.state[self.PARENT_TYPES] = set()
        self.state[self.RELATIONSHIPS] = set()
        self.state[self.CONNECTED] = set()
        self.state[self.ACTIONS] = dict()

        # Workaround for 401 error issue
        try:
            # Moves _translate_packages above translate_services to
            # make use of properties table in translate_services
            logger.debug("Murano grabbing packages")
            packages = self.murano_client.packages.list()
            self._translate_packages(packages)

            logger.debug("Murano grabbing environments")
            environments = self.murano_client.environments.list()
            self._translate_environments(environments)
            self._translate_services(environments)
            self._translate_deployments(environments)
            self._translate_connected()
        except murano_exceptions.HTTPException as e:
            raise e

    @classmethod
    def get_schema(cls):
        """Returns a dictionary of table schema.

        The dictionary mapping tablenames to the list of column names
        for that table. Both tablenames and columnnames are strings.
        """
        d = {}
        d[cls.OBJECTS] = ('object_id', 'owner_id', 'type')
        # parent_types include not only the type of object's immediate
        # parent but also all of its ancestors and its own type.  The
        # additional info helps writing better datalog rules.
        d[cls.PARENT_TYPES] = ('id', 'parent_type')
        d[cls.PROPERTIES] = ('owner_id', 'name', 'value')
        d[cls.RELATIONSHIPS] = ('source_id', 'target_id', 'name')
        d[cls.CONNECTED] = ('source_id', 'target_id')
        d[cls.STATES] = ('id', 'state')
        return d

    def _translate_environments(self, environments):
        """Translate the environments into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from environments
        """
        logger.debug("_translate_environments: %s", environments)
        if not environments:
            return
        self.state[self.STATES] = set()
        if self.OBJECTS not in self.state:
            self.state[self.OBJECTS] = set()
        if self.PROPERTIES not in self.state:
            self.state[self.PROPERTIES] = set()
        if self.PARENT_TYPES not in self.state:
            self.state[self.PARENT_TYPES] = set()
        if self.RELATIONSHIPS not in self.state:
            self.state[self.RELATIONSHIPS] = set()
        if self.CONNECTED not in self.state:
            self.state[self.CONNECTED] = set()

        env_type = 'io.murano.Environment'
        for env in environments:
            self.state[self.OBJECTS].add(
                (env.id, env.tenant_id, env_type))
            self.state[self.STATES].add((env.id, env.status))
            parent_types = self._get_parent_types(env_type)
            self._add_parent_types(env.id, parent_types)
            for key, value in env.to_dict().items():
                if key in self.UNUSED_ENV_PROPERTIES:
                    continue
                self._add_properties(env.id, key, value)

    def _translate_services(self, environments):
        """Translate the environment services into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from services
        """
        logger.debug("Murano grabbing environments services")
        if not environments:
            return
        for env in environments:
            services = self.murano_client.services.list(env.id)
            self._translate_environment_services(services, env.id)

    def _translate_environment_services(self, services, env_id):
        """Translate the environment services into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from services
        """

        # clean actions for given environment
        if self.ACTIONS not in self.state:
            self.state[self.ACTIONS] = dict()
        env_actions = self.state[self.ACTIONS][env_id] = set()

        if not services:
            return
        for s in services:
            s_dict = s.to_dict()
            s_id = s_dict['?']['id']
            s_type = s_dict['?']['type']
            self.state[self.OBJECTS].add((s_id, env_id, s_type))
            for key, value in s_dict.items():
                if key in ['instance', '?']:
                    continue
                self._add_properties(s_id, key, value)
                self._add_relationships(s_id, key, value)

            parent_types = self._get_parent_types(s_type)
            self._add_parent_types(s_id, parent_types)
            self._add_relationships(env_id, 'services', s_id)
            self._translate_service_action(s_dict, env_actions)

            if 'instance' not in s_dict:
                continue
            # populate service instance
            si_dict = s.instance
            si_id = si_dict['?']['id']
            si_type = si_dict['?']['type']
            self.state[self.OBJECTS].add((si_id, s_id, si_type))

            for key, value in si_dict.items():
                if key in ['?']:
                    continue
                self._add_properties(si_id, key, value)
                if key not in ['image']:
                    # there's no murano image object in the environment,
                    # therefore glance 'image' relationship is irrelevant
                    # at this point.
                    self._add_relationships(si_id, key, value)
            # There's a relationship between the service and instance
            self._add_relationships(s_id, 'instance', si_id)

            parent_types = self._get_parent_types(si_type)
            self._add_parent_types(si_id, parent_types)
            self._translate_service_action(si_dict, env_actions)

    def _translate_service_action(self, obj_dict, env_actions):
        """Translates environment's object actions to env_actions structure.

        env_actions: [(obj_id, action_id, action_name, enabled)]
        :param obj_dict: object dictionary
        :param env_actions: set of environment actions
        """
        obj_id = obj_dict['?']['id']
        if '_actions' in obj_dict['?']:
            o_actions = obj_dict['?']['_actions']
            if not o_actions:
                return
            for action_id, action_value in o_actions.items():
                action_name = action_value.get('name', '')
                enabled = action_value.get('enabled', False)
                action = (obj_id, action_id, action_name, enabled)
                env_actions.add(action)
                # TODO(tranldt): support action arguments.
                #  If action arguments are included in '_actions',
                #  they can be populated into tables.

    def _translate_deployments(self, environments):
        """Translate the environment deployments into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from deployments
        """
        if not environments:
            return
        for env in environments:
            deployments = self.murano_client.deployments.list(env.id)
            self._translate_environment_deployments(deployments, env.id)

    def _translate_environment_deployments(self, deployments, env_id):
        """Translate the environment deployments into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from deployments
        """
        if not deployments:
            return
        for d in deployments:
            if 'defaultNetworks' not in d.description:
                continue
            default_networks = d.description['defaultNetworks']
            net_id = None
            if 'environment' in default_networks:
                net_id = default_networks['environment']['?']['id']
                net_type = default_networks['environment']['?']['type']
                self.state[self.OBJECTS].add((net_id, env_id, net_type))

                parent_types = self._get_parent_types(net_type)
                self._add_parent_types(net_id, parent_types)

                for key, value in default_networks['environment'].items():
                    if key in ['?']:
                        continue
                    self._add_properties(net_id, key, value)

            if not net_id:
                continue
            self._add_relationships(env_id, 'defaultNetworks', net_id)
            for key, value in default_networks.items():
                if key in ['environment']:
                    # data from environment already populated
                    continue
                new_key = 'defaultNetworks.' + key
                self._add_properties(net_id, new_key, value)
            # services from deployment are not of interest because the same
            # info is obtained from services API

    def _translate_packages(self, packages):
        """Translate the packages into tables.

        Assigns self.state[tablename] for all those TABLENAMEs
        generated from packages/applications
        """
        # packages is a generator type
        if not packages:
            return
        if self.OBJECTS not in self.state:
            self.state[self.OBJECTS] = set()
        if self.PROPERTIES not in self.state:
            self.state[self.PROPERTIES] = set()

        for pkg in packages:
            logger.debug("pkg=%s", pkg.to_dict())
            pkg_type = pkg.type
            if pkg.type == 'Application':
                pkg_type = 'io.murano.Application'
            self.state[self.OBJECTS].add((pkg.id, pkg.owner_id, pkg_type))

            for key, value in pkg.to_dict().items():
                if key in self.UNUSED_PKG_PROPERTIES:
                    continue
                self._add_properties(pkg.id, key, value)

    def _add_properties(self, obj_id, key, value):
        """Add a set of (obj_id, key, value) to properties table.

        :param obj_id: uuid of object
        :param key: property name. For the case value is a list, the
        same key is used for multiple values.
        :param value: property value. If value is a dict, the nested
        properties will be mapped using dot notation.
        """
        if value is None or value == '':
            return
        if isinstance(value, dict):
            for k, v in value.items():
                new_key = key + "." + k
                self._add_properties(obj_id, new_key, v)
        elif isinstance(value, list):
            if len(value) == 0:
                return
            for item in value:
                self.state[self.PROPERTIES].add(
                    (obj_id, key, utils.value_to_congress(item)))
        else:
            self.state[self.PROPERTIES].add(
                (obj_id, key, utils.value_to_congress(value)))

    def _add_relationships(self, obj_id, key, value):
        """Add a set of (obj_id, value, key) to relationships table.

        :param obj_id: source uuid
        :param key: relationship name
        :param value: target uuid
        """
        if (not isinstance(value, six.string_types) or
                not uuidutils.is_uuid_like(value)):
            return
        logger.debug("Relationship: source = %s, target = %s, rel_name = %s"
                     % (obj_id, value, key))
        self.state[self.RELATIONSHIPS].add((obj_id, value, key))

    def _transitive_closure(self):
        """Computes transitive closure on a directed graph.

        In other words computes reachability within the graph.
        E.g. {(1, 2), (2, 3)} -> {(1, 2), (2, 3), (1, 3)}
        (1, 3) was added because there is path from 1 to 3 in the graph.
        """
        closure = self.state[self.CONNECTED]
        while True:
            # Attempts to discover new transitive relations
            # by joining 2 subsequent relations/edges within the graph.
            new_relations = {(x, w) for x, y in closure
                             for q, w in closure if q == y}
            # Creates union with already discovered relations.
            closure_until_now = closure | new_relations
            # If no new relations were discovered in last cycle
            # the computation is finished.
            if closure_until_now == closure:
                self.state[self.CONNECTED] = closure
                break
            closure = closure_until_now

    def _add_connected(self, source_id, target_id):
        """Looks up the target_id in objects and add links to connected table.

        Adds sets of (source_id, target_id) to connected table along
        with its indirections.
        :param source_id: source uuid
        :param target_id: target uuid
        """
        for row in self.state[self.OBJECTS]:
            if row[1] == target_id:
                self.state[self.CONNECTED].add((row[1], row[0]))
                self.state[self.CONNECTED].add((source_id, row[0]))
        self.state[self.CONNECTED].add((source_id, target_id))

    def _translate_connected(self):
        """Translates relationships table into connected table."""
        for row in self.state[self.RELATIONSHIPS]:
            self._add_connected(row[0], row[1])
        self._transitive_closure()

    def _add_parent_types(self, obj_id, parent_types):
        """Add sets of (obj_id, parent_type) to parent_types table.

        :param obj_id: uuid of object
        :param parent_types: list of parent type string
        """
        if parent_types:
            for p_type in parent_types:
                self.state[self.PARENT_TYPES].add((obj_id, p_type))

    def _get_package_type(self, class_name):
        """Determine whether obj_type is an Application or Library.

        :param class_name: <string> service/application class name
            e.g. io.murano.apps.linux.Telnet.
        :return - package type (e.g. 'Application') if found.
            - None if no package type found.
        """
        pkg_type = None
        if self.PROPERTIES in self.state:
            idx_uuid = 0
            idx_value = 2
            uuid = None
            for row in self.state[self.PROPERTIES]:
                if 'class_definitions' in row and class_name in row:
                    uuid = row[idx_uuid]
                    break
            if uuid:
                for row in self.state[self.PROPERTIES]:
                    if 'type' in row and uuid == row[idx_uuid]:
                        pkg_type = row[idx_value]

        # If the package is removed after deployed, its properties
        #  are not known and so above search will fail. In that case
        #  let's check for class_name prefix as the last resort.
        if not pkg_type:
            for prefix in self.APPS_TYPE_PREFIXES:
                if prefix in class_name:
                    pkg_type = 'Application'
                    break
        return pkg_type

    def _get_parent_types(self, obj_type):
        """Get class types of all OBJ_TYPE's parents including itself.

        Look up the hierarchy of OBJ_TYPE and return types of all its
        ancestor including its own type.
        :param obj_type: <string>
        """
        class_types = []
        p = lambda x: inspect.isclass(x)
        g = inspect.getmembers(murano_classes, p)
        for name, cls in g:
            logger.debug("%s: %s" % (name, cls))
            if (cls is murano_classes.IOMuranoApps and
                    self._get_package_type(obj_type) == 'Application'):
                cls.name = obj_type
            if 'get_parent_types' in dir(cls):
                class_types = cls.get_parent_types(obj_type)
                if class_types:
                    break
        return class_types

    def _call_murano_action(self, environment_id, object_id, action_name):
        """Invokes action of object in Murano environment.

        :param environment_id: uuid
        :param object_id: uuid
        :param action_name: string
        """
        # get action id using object_id, env_id and action name
        logger.debug("Requested Murano action invoke %s on %s in %s",
                     action_name, object_id, environment_id)
        if (not self.state[self.ACTIONS] or
                environment_id not in self.state[self.ACTIONS]):
            logger.warning('Datasource "%s" found no actions for '
                           'environment "%s"', self.name, environment_id)
            return
        env_actions = self.state[self.ACTIONS][environment_id]
        for env_action in env_actions:
            ea_obj_id, ea_action_id, ea_action_name, ea_enabled = env_action
            if (object_id == ea_obj_id and action_name == ea_action_name
                    and ea_enabled):
                logger.debug("Invoking Murano action_id = %s, action_name %s",
                             ea_action_id, ea_action_name)
                # TODO(tranldt): support action arguments
                task_id = self.murano_client.actions.call(environment_id,
                                                          ea_action_id)
                logger.debug("Murano action invoked %s - task id %s",
                             ea_action_id, task_id)
                self.action_call_returns.append(task_id)

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        logger.info("%s:: executing %s on %s", self.name, action, action_args)
        self.action_call_returns = []
        positional_args = action_args.get('positional', [])
        logger.debug('Processing action execution: action = %s, '
                     'positional args = %s', action, positional_args)
        try:
            env_id = positional_args[0]
            obj_id = positional_args[1]
            action_name = positional_args[2]
            self._call_murano_action(env_id, obj_id, action_name)
        except Exception as e:
            logger.exception(e.message)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_datasources

Revision ID: 3cee191c4f84
Revises: 56e86d51ec62
Create Date: 2015-02-05 13:30:04.272571

"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# revision identifiers, used by Alembic.
revision = '3cee191c4f84'
down_revision = '56e86d51ec62'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'datasources',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('driver', sa.String(length=255), nullable=True),
        sa.Column('config', sa.Text(), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('enabled', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name'),
        mysql_engine='InnoDB')


def downgrade():
    op.drop_table('datasources')



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json
import os

import fixtures
from oslo_config import cfg

import congress.common.policy
from congress.tests import fake_policy

CONF = cfg.CONF


class PolicyFixture(fixtures.Fixture):

    def setUp(self):
        super(PolicyFixture, self).setUp()
        self.policy_dir = self.useFixture(fixtures.TempDir())
        self.policy_file_name = os.path.join(self.policy_dir.path,
                                             'policy.json')
        with open(self.policy_file_name, 'w') as policy_file:
            policy_file.write(fake_policy.policy_data)
        CONF.set_override('policy_file', self.policy_file_name, 'oslo_policy')
        congress.common.policy.reset()
        congress.common.policy.init()
        self.addCleanup(congress.common.policy.reset)


class RoleBasedPolicyFixture(fixtures.Fixture):

    def __init__(self, role="admin", *args, **kwargs):
        super(RoleBasedPolicyFixture, self).__init__(*args, **kwargs)
        self.role = role

    def setUp(self):
        """Set up the env for test.

        Copy live policy.json file and convert all actions to
        allow users of the specified role only.
        """
        super(RoleBasedPolicyFixture, self).setUp()
        policy = json.load(open(CONF.oslo_policy.policy_file))

        # Convert all actions to require specified role
        for action, rule in policy.items():
            policy[action] = 'role:%s' % self.role

        self.policy_dir = self.useFixture(fixtures.TempDir())
        self.policy_file_name = os.path.join(self.policy_dir.path,
                                             'policy.json')
        with open(self.policy_file_name, 'w') as policy_file:
            json.dump(policy, policy_file)
        CONF.set_override('policy_file', self.policy_file_name, 'oslo_policy')
        congress.common.policy.reset()
        congress.common.policy.init()
        self.addCleanup(congress.common.policy.reset)



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import contextlib
import logging as python_logging

from oslo_log import log as logging
import six

from congress.datalog import utility
from congress.tests import base

LOG = logging.getLogger(__name__)


class TestGraph(base.TestCase):
    def test_nodes(self):
        """Test addition/deletion of nodes."""
        g = utility.Graph()
        g.add_node(1)
        g.add_node(2)
        g.add_node(3)
        g.add_node(1)
        self.assertTrue(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 3)

        g.delete_node(1)
        self.assertFalse(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 2)

        g.delete_node(1)
        self.assertFalse(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 2)

        g.add_node(1)
        self.assertTrue(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 3)

        g.delete_node(2)
        self.assertTrue(g.node_in(1))
        self.assertFalse(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 2)

    def test_edges(self):
        """Test addition/deletion of edges."""
        g = utility.Graph()

        g.add_edge(1, 2)
        g.add_edge(2, 3)
        g.add_edge(2, 4)
        g.add_edge(1, 2)
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 7)

        g.delete_edge(2, 4)
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertFalse(g.edge_in(2, 4))
        self.assertEqual(len(g), 6)

        g.delete_edge(2, 3)
        self.assertTrue(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(2, 3))
        self.assertFalse(g.edge_in(2, 4))
        self.assertEqual(len(g), 5)

        g.delete_edge(2, 3)
        self.assertTrue(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(2, 3))
        self.assertFalse(g.edge_in(2, 4))
        self.assertEqual(len(g), 5)

        g.add_edge(2, 3)
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertFalse(g.edge_in(2, 4))
        self.assertEqual(len(g), 6)

    def test_or(self):
        g1 = utility.Graph()
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)

        g2 = utility.Graph()
        g2.add_edge(2, 3)
        g2.add_edge(3, 4)

        g = g1 | g2
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(3, 4))
        self.assertIsInstance(g, utility.Graph)
        self.assertTrue(g1.edge_in(1, 2))
        self.assertTrue(g1.edge_in(2, 3))
        self.assertFalse(g1.edge_in(3, 4))
        self.assertFalse(g2.edge_in(1, 2))
        self.assertTrue(g2.edge_in(2, 3))
        self.assertTrue(g2.edge_in(3, 4))

    def test_ior(self):
        g1 = utility.Graph()
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)

        g2 = utility.Graph()
        g2.add_edge(2, 3)
        g2.add_edge(3, 4)

        g1 |= g2
        self.assertTrue(g1.edge_in(1, 2))
        self.assertTrue(g1.edge_in(2, 3))
        self.assertTrue(g1.edge_in(3, 4))
        self.assertFalse(g2.edge_in(1, 2))
        self.assertTrue(g2.edge_in(2, 3))
        self.assertTrue(g2.edge_in(3, 4))

    def test_cycle(self):
        g1 = utility.Graph()
        g1.add_edge(1, 2)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(2, 3)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(2, 4)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(4, 1)
        self.assertTrue(g1.has_cycle())
        g1.delete_edge(2, 3)
        self.assertTrue(g1.has_cycle())
        g1.delete_edge(2, 4)
        self.assertFalse(g1.has_cycle())

        actual_cycle_as_set = frozenset(utility.Cycle(['p', 'q', 't', 'p']))
        expected_cycle_as_set = frozenset([('p', 'q'), ('q', 't'), ('t', 'p')])
        self.assertEqual(actual_cycle_as_set, expected_cycle_as_set)

        g = utility.Graph()
        g.add_edge('p', 'q')
        g.add_edge('p', 'r')
        g.add_edge('q', 't')
        g.add_edge('q', 's')
        g.add_edge('t', 't')
        g.add_edge('t', 'p')
        g.add_edge('t', 'q')
        self.assertTrue(g.has_cycle())
        self.assertEqual(len(g.cycles()), 3)
        expected_cycle_set = set([
            utility.Cycle(['p', 'q', 't', 'p']),
            utility.Cycle(['q', 't', 'q']),
            utility.Cycle(['t', 't'])
        ])
        actual_cycle_set = set([
            utility.Cycle(g.cycles()[0]),
            utility.Cycle(g.cycles()[1]),
            utility.Cycle(g.cycles()[2])
        ])
        self.assertEqual(expected_cycle_set, actual_cycle_set)

    def test_find_reachable_nodes(self):
        g1 = utility.Graph()
        self.assertEqual(g1.find_reachable_nodes([1]), set())
        g1.add_edge(0, 1)
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)
        g1.add_edge(2, 4)
        g1.add_edge(3, 5)
        g1.add_edge(0, 6)
        g1.add_edge(7, 8)
        g1.add_edge(8, 9)
        g1.add_edge(10, 5)
        g1.add_edge(11, 12)
        self.assertEqual(g1.find_reachable_nodes([]), set())
        self.assertEqual(g1.find_reachable_nodes([1]), set([1, 2, 3, 4, 5]))
        self.assertEqual(g1.find_reachable_nodes([10]), set([10, 5]))
        self.assertEqual(g1.find_reachable_nodes([1, 10]),
                         set([1, 2, 3, 4, 5, 10]))
        self.assertEqual(g1.find_reachable_nodes([5]), set([5]))
        self.assertEqual(g1.find_reachable_nodes([11]), set([11, 12]))
        g1.add_edge(5, 2)
        self.assertEqual(g1.find_reachable_nodes([10]), set([10, 5, 2, 3, 4]))

    def test_dependencies(self):
        g1 = utility.Graph()
        self.assertIsNone(g1.dependencies(1))
        g1.add_edge(0, 1)
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)
        g1.add_edge(2, 4)
        g1.add_edge(3, 5)
        g1.add_edge(0, 6)
        g1.add_edge(7, 8)
        g1.add_edge(8, 9)
        g1.add_edge(10, 5)
        g1.add_edge(11, 12)
        self.assertEqual(g1.dependencies(1), set([1, 2, 3, 4, 5]))
        self.assertEqual(g1.dependencies(10), set([10, 5]))
        self.assertEqual(g1.dependencies(5), set([5]))
        self.assertEqual(g1.dependencies(11), set([11, 12]))

    def test_cyclic_dependencies(self):
        g1 = utility.Graph()
        self.assertIsNone(g1.dependencies(1))
        g1.add_edge(0, 1)
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)
        g1.add_edge(2, 4)
        g1.add_edge(3, 5)
        g1.add_edge(0, 6)
        g1.add_edge(7, 8)
        g1.add_edge(8, 9)
        g1.add_edge(10, 5)
        g1.add_edge(11, 12)
        g1.add_edge(5, 2)  # create cycle
        self.assertEqual(g1.dependencies(1), set([1, 2, 3, 4, 5]))
        self.assertEqual(g1.dependencies(10), set([10, 5, 2, 3, 4]))
        self.assertEqual(g1.dependencies(5), set([5, 2, 3, 4]))
        self.assertEqual(g1.dependencies(11), set([11, 12]))

    def test_find_dependent_nodes(self):
        g1 = utility.Graph()
        self.assertEqual(g1.find_dependent_nodes([1]), set([1]))
        g1.add_edge(0, 1)
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)
        g1.add_edge(2, 4)
        g1.add_edge(3, 5)
        g1.add_edge(0, 6)
        g1.add_edge(7, 8)
        g1.add_edge(8, 9)
        g1.add_edge(10, 5)
        g1.add_edge(11, 12)
        self.assertEqual(g1.find_dependent_nodes([0]), set([0]))
        self.assertEqual(g1.find_dependent_nodes([2]), set([2, 1, 0]))
        self.assertEqual(g1.find_dependent_nodes([5]),
                         set([5, 0, 1, 2, 3, 10]))
        self.assertEqual(g1.find_dependent_nodes([12]), set([11, 12]))
        self.assertEqual(g1.find_dependent_nodes([5, 6]),
                         set([5, 0, 1, 2, 3, 10, 6]))
        g1.add_edge(5, 2)  # add cycle
        self.assertEqual(g1.find_dependent_nodes([2]),
                         set([5, 0, 1, 2, 3, 10]))


class TestBagGraph(base.TestCase):
    def test_nodes(self):
        """Test addition/deletion of nodes."""
        g = utility.BagGraph()
        g.add_node(1)
        g.add_node(2)
        g.add_node(3)
        g.add_node(1)
        self.assertTrue(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 4)

        g.delete_node(1)
        self.assertTrue(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 3)

        g.delete_node(1)
        self.assertFalse(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 2)

        g.add_node(1)
        self.assertTrue(g.node_in(1))
        self.assertTrue(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 3)

        g.delete_node(2)
        self.assertTrue(g.node_in(1))
        self.assertFalse(g.node_in(2))
        self.assertTrue(g.node_in(3))
        self.assertEqual(len(g), 2)

    def test_edges(self):
        """Test addition/deletion of edges."""
        g = utility.BagGraph()

        g.add_edge(1, 2)
        g.add_edge(2, 3)
        g.add_edge(2, 4)
        g.add_edge(1, 2)
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 12)

        g.delete_edge(1, 2)
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 9)

        g.delete_edge(1, 2)
        self.assertFalse(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 6)

        g.delete_edge(2, 3)
        self.assertFalse(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 3)

        g.add_edge(1, 2)
        self.assertTrue(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(2, 4))
        self.assertEqual(len(g), 6)

        g.add_node(1)
        self.assertEqual(g.node_count(1), 2)

    def test_edge_labels(self):
        g = utility.BagGraph()

        g.add_edge(1, 2)
        g.add_edge(1, 2, "label1")
        g.add_edge(1, 2, "label1")
        g.add_edge(1, 2, "label2")
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(1, 2, "label1"))
        self.assertTrue(g.edge_in(1, 2, "label2"))
        self.assertFalse(g.edge_in(1, 2, "non-existent"))
        self.assertEqual(g.edge_count(1, 2, "label1"), 2)

        g.delete_edge(1, 2, "label1")
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(1, 2, "label1"))
        self.assertTrue(g.edge_in(1, 2, "label2"))
        self.assertFalse(g.edge_in(1, 2, "non-existent"))
        self.assertEqual(g.edge_count(1, 2, "label1"), 1)

        g.delete_edge(1, 2, "label1")
        self.assertTrue(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(1, 2, "label1"))
        self.assertTrue(g.edge_in(1, 2, "label2"))
        self.assertFalse(g.edge_in(1, 2, "non-existent"))
        self.assertEqual(g.edge_count(1, 2, "label1"), 0)

        g.delete_edge(1, 2, "label2")
        self.assertTrue(g.edge_in(1, 2))
        self.assertFalse(g.edge_in(1, 2, "label1"))
        self.assertFalse(g.edge_in(1, 2, "label2"))
        self.assertFalse(g.edge_in(1, 2, "non-existent"))
        self.assertEqual(g.edge_count(1, 2, "label1"), 0)

    def test_or(self):
        g1 = utility.BagGraph()
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)

        g2 = utility.BagGraph()
        g2.add_edge(2, 3)
        g2.add_edge(3, 4)

        g = g1 | g2
        self.assertTrue(g.edge_in(1, 2))
        self.assertTrue(g.edge_in(2, 3))
        self.assertTrue(g.edge_in(3, 4))
        self.assertEqual(g.edge_count(2, 3), 2)
        self.assertIsInstance(g, utility.Graph)
        self.assertTrue(g1.edge_in(1, 2))
        self.assertTrue(g1.edge_in(2, 3))
        self.assertFalse(g1.edge_in(3, 4))
        self.assertEqual(g1.edge_count(2, 3), 1)
        self.assertFalse(g2.edge_in(1, 2))
        self.assertTrue(g2.edge_in(2, 3))
        self.assertTrue(g2.edge_in(3, 4))
        self.assertEqual(g2.edge_count(2, 3), 1)

    def test_ior(self):
        g1 = utility.BagGraph()
        g1.add_edge(1, 2)
        g1.add_edge(2, 3)

        g2 = utility.BagGraph()
        g2.add_edge(2, 3)
        g2.add_edge(3, 4)

        g1 |= g2
        self.assertTrue(g1.edge_in(1, 2))
        self.assertTrue(g1.edge_in(2, 3))
        self.assertTrue(g1.edge_in(3, 4))
        self.assertIsInstance(g1, utility.BagGraph)
        self.assertEqual(g1.edge_count(2, 3), 2)
        self.assertFalse(g2.edge_in(1, 2))
        self.assertTrue(g2.edge_in(2, 3))
        self.assertTrue(g2.edge_in(3, 4))
        self.assertEqual(g2.edge_count(2, 3), 1)

    def test_cycle(self):
        g1 = utility.BagGraph()
        g1.add_edge(1, 2)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(2, 3)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(2, 4)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(2, 4)
        self.assertFalse(g1.has_cycle())
        g1.add_edge(4, 1)
        self.assertTrue(g1.has_cycle())
        g1.delete_edge(2, 3)
        self.assertTrue(g1.has_cycle())
        g1.delete_edge(2, 4)
        self.assertTrue(g1.has_cycle())
        g1.delete_edge(2, 4)
        self.assertFalse(g1.has_cycle())


class TestIterstr(base.TestCase):
    class X(object):
        def __init__(self, v):
            self.v = v

        def __str__(self):
            return "%s" % self.v

        def __repr__(self):
            return "X:%s" % self.v

    @contextlib.contextmanager
    def get_logging_fixtures(self):
        stream = six.moves.StringIO()
        handler = python_logging.StreamHandler(stream)
        try:
            logger = python_logging.getLogger(self.__class__.__name__)
            logger.setLevel(python_logging.INFO)
            handler.setLevel(python_logging.INFO)
            logger.addHandler(handler)
            try:
                yield (stream, handler, logger)
            finally:
                logger.removeHandler(handler)
        finally:
            handler.close()

    def test__str__returns_informal_representation(self):
        xs = map(TestIterstr.X, range(5))
        observed = utility.iterstr(xs)
        self.assertEqual("[0;1;2;3;4]", str(observed))
        self.assertEqual("[0;1;2;3;4]", "{}".format(observed))
        self.assertEqual("[0;1;2;3;4]", "%s" % observed)

    def test__repr__returns_formal_representation(self):
        xs = map(TestIterstr.X, range(5))
        observed = utility.iterstr(xs)
        self.assertEqual("[X:0;X:1;X:2;X:3;X:4]", repr(observed))
        self.assertEqual("[X:0;X:1;X:2;X:3;X:4]", "{!r}".format(observed))
        self.assertEqual("[X:0;X:1;X:2;X:3;X:4]", "%r" % observed)

    def test_empty(self):
        xs = map(TestIterstr.X, range(0))
        observed = utility.iterstr(xs)
        self.assertEqual("[]", str(observed))
        self.assertEqual("[]", repr(observed))

    def test_logging_basic_integration(self):
        with self.get_logging_fixtures() as (stream, handler, logger):
            iterable = utility.iterstr(map(TestIterstr.X, range(5)))
            logger.info("some message %s", iterable)
            handler.flush()
            self.assertEqual("some message [0;1;2;3;4]\n", stream.getvalue())

    def test_logging_skips_interpolation(self):
        with self.get_logging_fixtures() as (stream, handler, logger):
            iterable = utility.iterstr(map(TestIterstr.X, range(5)))
            logger.debug("some message %s", iterable)
            self.assertIsNone(iterable._str_interp)



# Copyright (c) 2016 Styra, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress.tests.datalog import test_compiler
from congress.tests.datalog import test_factset
from congress.tests.datalog import test_materialized
from congress.tests.datalog import test_nonrecur
from congress.tests.datalog import test_ordered_set
from congress.tests.datalog import test_ruleset
from congress.tests.datalog import test_unify
from congress.tests.datalog import test_utility


class TestCompilerParser(test_compiler.TestParser):
    pass


class TestCompilerColumn(test_compiler.TestColumnReferences):
    pass


class TestCompilerCompiler(test_compiler.TestCompiler):
    pass


class TestCompilerGraph(test_compiler.TestDependencyGraph):
    pass


class TestFact(test_factset.TestFactSet):
    pass


class TestMaterialized(test_materialized.TestRuntime):
    pass


class TestNonrecurRuntime(test_nonrecur.TestRuntime):
    pass


class TestNonrecurNegation(test_nonrecur.TestSelectNegation):
    pass


class TestNonrecurArity(test_nonrecur.TestArity):
    pass


class TestNonrecurInstances(test_nonrecur.TestInstances):
    pass


class TestOrdered(test_ordered_set.TestOrderedSet):
    pass


class TestRuleSet(test_ruleset.TestRuleSet):
    pass


class TestUnifyUnify(test_unify.TestUnify):
    pass


class TestUnifyMatch(test_unify.TestMatch):
    pass


class TestUtilityGraph(test_utility.TestGraph):
    pass


class TestUtilityBagGraph(test_utility.TestBagGraph):
    pass


class TestUtilityIterstr(test_utility.TestIterstr):
    pass



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json
import six
import traceback
import uuid

import eventlet
eventlet.monkey_patch()

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging import exceptions as messaging_exceptions
from oslo_utils import importutils
from oslo_utils import uuidutils

from congress.datasources import constants
from congress.db import api as db
from congress.db import datasources as datasources_db
from congress.dse2.control_bus import DseNodeControlBus
from congress import exception


LOG = logging.getLogger()


_dse_opts = [
    cfg.StrOpt('node_id', help='Unique ID of this DseNode on the DSE')
]
cfg.CONF.register_opts(_dse_opts, group='dse')


class DseNode(object):
    """Addressable entity participating on the DSE message bus.

    The Data Services Engine (DSE) is comprised of one or more DseNode
    instances that each may run one or more DataService instances.  All
    communication between data services uses the DseNode interface.

    Attributes:
        node_id: The unique ID of this node on the DSE.
        messaging_config: Configuration options for the message bus.  See
                          oslo.messaging for more details.
        node_rpc_endpoints: List of object instances exposing a remotely
                            invokable interface.
    """
    RPC_VERSION = '1.0'
    CONTROL_TOPIC = 'congress-control'
    SERVICE_TOPIC_PREFIX = 'congress-service-'

    def node_rpc_target(self, namespace=None, server=None, fanout=False):
        return messaging.Target(topic=self._add_partition(self.CONTROL_TOPIC),
                                version=self.RPC_VERSION,
                                namespace=namespace,
                                server=server,
                                fanout=fanout)

    def service_rpc_target(self, service_id, namespace=None, server=None,
                           fanout=False):
        topic = self._add_partition(self.SERVICE_TOPIC_PREFIX + service_id)
        return messaging.Target(topic=topic,
                                version=self.RPC_VERSION,
                                namespace=namespace,
                                server=server,
                                fanout=fanout)

    def _add_partition(self, topic, partition_id=None):
        """Create a seed-specific version of an oslo-messaging topic."""
        partition_id = partition_id or self.partition_id
        if partition_id is None:
            return topic
        return topic + "-" + str(partition_id)

    def __init__(self, messaging_config, node_id, node_rpc_endpoints,
                 partition_id=None):
        self.messaging_config = messaging_config
        self.node_id = node_id
        self.node_rpc_endpoints = node_rpc_endpoints
        # unique identifier shared by all nodes that can communicate
        self.partition_id = partition_id
        self.node_rpc_endpoints.append(DseNodeEndpoints(self))
        self._running = False
        self._services = []
        self.instance = uuid.uuid4()
        self.context = self._message_context()
        self.transport = messaging.get_transport(
            self.messaging_config,
            allowed_remote_exmods=[exception.__name__, ])
        self._rpctarget = self.node_rpc_target(self.node_id, self.node_id)
        self._rpcserver = messaging.get_rpc_server(
            self.transport, self._rpctarget, self.node_rpc_endpoints,
            executor='eventlet')
        self._service_rpc_servers = {}  # {service_id => (rpcserver, target)}

        # # keep track of what publisher/tables local services subscribe to
        # subscribers indexed by publisher and table:
        # {publisher_id ->
        #     {table_name -> set_of_subscriber_ids}}
        self.subscriptions = {}

        # Note(ekcs): A little strange that _control_bus starts before self?
        self._control_bus = DseNodeControlBus(self)
        self.register_service(self._control_bus)
        # load configured drivers
        self.loaded_drivers = self.load_drivers()
        self.start()

    def __del__(self):
        self.stop()
        self.wait()

    def __repr__(self):
        return self.__class__.__name__ + "<%s>" % self.node_id

    def _message_context(self):
        return {'node_id': self.node_id, 'instance': str(self.instance)}

    def register_service(self, service, index=None):
        assert service.node is None
        service.node = self
        if index is not None:
            self._services.insert(index, service)
        else:
            self._services.append(service)

        target = self.service_rpc_target(service.service_id,
                                         server=self.node_id)
        srpc = messaging.get_rpc_server(
            self.transport, target, service.rpc_endpoints(),
            executor='eventlet')
        self._service_rpc_servers[service.service_id] = (srpc, target)
        service.start()
        srpc.start()
        LOG.debug('<%s> Service %s RPC Server listening on %s',
                  self.node_id, service.service_id, target)

    def unregister_service(self, service_id, index=None):
        self._services = [s for s in self._services
                          if s.service_id != service_id]
        srpc, _ = self._service_rpc_servers[service_id]
        srpc.stop()
        srpc.wait()
        del self._service_rpc_servers[service_id]

    def get_services(self, hidden=False):
        """Return all local service objects."""
        if hidden:
            return self._services
        return [s for s in self._services if s.service_id[0] != '_']

    def get_global_service_names(self, hidden=False):
        """Return names of all services on all nodes."""
        services = self.get_services(hidden=hidden)
        local_services = [s.service_id for s in services]
        # Also, check services registered on other nodes
        peer_nodes = self.dse_status()['peers']
        peer_services = []
        for node in peer_nodes.values():
            peer_services.extend(
                [srv['service_id'] for srv in node['services']])
        return set(local_services + peer_services)

    def service_object(self, name):
        """Returns the service object of the given name.  None if not found."""
        for s in self._services:
            if s.service_id == name:
                return s

    def start(self):
        LOG.debug("<%s> DSE Node '%s' starting with %s sevices...",
                  self.node_id, self.node_id, len(self._services))

        # Start Node RPC server
        self._rpcserver.start()
        LOG.debug('<%s> Node RPC Server listening on %s',
                  self.node_id, self._rpctarget)

        # Start Service RPC server(s)
        for s in self._services:
            s.start()
            sspec = self._service_rpc_servers.get(s.service_id)
            assert sspec is not None
            srpc, target = sspec
            srpc.start()
            LOG.debug('<%s> Service %s RPC Server listening on %s',
                      self.node_id, s.service_id, target)

        self._running = True

    def stop(self):
        LOG.info("Stopping DSE node '%s'" % self.node_id)
        for srpc, target in self._service_rpc_servers.values():
            srpc.stop()
        for s in self._services:
            s.stop()
        self._rpcserver.stop()
        self._running = False

    def wait(self):
        for s in self._services:
            s.wait()
        self._rpcserver.wait()

    def dse_status(self):
        """Return latest observation of DSE status."""
        return self._control_bus.dse_status()

    def is_valid_service(self, service_id):
        return service_id in self.get_global_service_names(hidden=True)

    def invoke_node_rpc(self, node_id, method, **kwargs):
        """Invoke RPC method on a DSE Node.

        Args:
            node_id: The ID of the node on which to invoke the call.
            method: The method name to call.
            kwargs: A dict of method arguments.

        Returns:
            The result of the method invocation.

        Raises: MessagingTimeout, RemoteError, MessageDeliveryFailure
        """
        target = self.node_rpc_target(server=node_id)
        LOG.trace("<%s> Invoking RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        return client.call(self.context, method, **kwargs)

    def broadcast_node_rpc(self, method, **kwargs):
        """Invoke RPC method on all DSE Nodes.

        Args:
            method: The method name to call.
            kwargs: A dict of method arguments.

        Returns:
            None - Methods are invoked asynchronously and results are dropped.

        Raises: RemoteError, MessageDeliveryFailure
        """
        target = self.node_rpc_target(fanout=True)
        LOG.trace("<%s> Casting RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        client.cast(self.context, method, **kwargs)

    def invoke_service_rpc(self, service_id, method, **kwargs):
        """Invoke RPC method on a DSE Service.

        Args:
            service_id: The ID of the data service on which to invoke the call.
            method: The method name to call.
            kwargs: A dict of method arguments.

        Returns:
            The result of the method invocation.

        Raises: MessagingTimeout, RemoteError, MessageDeliveryFailure, NotFound
        """
        target = self.service_rpc_target(service_id)
        LOG.trace("<%s> Invoking RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        # Using the control bus to check if the service exists before
        #   running the RPC doesn't always work, either because of bugs
        #   or nondeterminism--not clear which.
        try:
            result = client.call(self.context, method, **kwargs)
        except messaging_exceptions.MessagingTimeout:
            msg = "service '%s' could not be found"
            raise exception.NotFound(msg % service_id)
        LOG.trace("<%s> RPC call returned: %s", self.node_id, result)
        return result

    def broadcast_service_rpc(self, service_id, method, **kwargs):
        """Invoke RPC method on all insances of service_id.

        Args:
            service_id: The ID of the data service on which to invoke the call.
            method: The method name to call.
            kwargs: A dict of method arguments.

        Returns:
            None - Methods are invoked asynchronously and results are dropped.

        Raises: RemoteError, MessageDeliveryFailure
        """
        if not self.is_valid_service(service_id):
            msg = "service '%s' is not a registered service"
            raise exception.NotFound(msg % service_id)

        target = self.service_rpc_target(service_id, fanout=True)
        LOG.trace("<%s> Casting RPC '%s' on %s", self.node_id, method, target)
        client = messaging.RPCClient(self.transport, target)
        client.cast(self.context, method, **kwargs)

    def publish_table(self, publisher, table, data):
        """Invoke RPC method on all insances of service_id.

        Args:
            service_id: The ID of the data service on which to invoke the call.
            method: The method name to call.
            kwargs: A dict of method arguments.

        Returns:
            None - Methods are invoked asynchronously and results are dropped.

        Raises: RemoteError, MessageDeliveryFailure
        """
        LOG.trace("<%s> Publishing from '%s' table %s: %s",
                  self.node_id, publisher, table, data)
        self.broadcast_node_rpc("handle_publish", publisher=publisher,
                                table=table, data=data)

    def table_subscribers(self, publisher, table):
        """List services on this node that subscribes to publisher/table."""
        return self.subscriptions.get(
            publisher, {}).get(table, [])

    def subscribe_table(self, subscriber, publisher, table):
        """Prepare local service to receives publications from target/table."""
        # data structure: {service -> {target -> set-of-tables}
        LOG.trace("subscribing %s to %s:%s", subscriber, publisher, table)
        if publisher not in self.subscriptions:
            self.subscriptions[publisher] = {}
        if table not in self.subscriptions[publisher]:
            self.subscriptions[publisher][table] = set()
        self.subscriptions[publisher][table].add(subscriber)

        snapshot = self.invoke_service_rpc(
            publisher, "get_snapshot", table=table)

        # oslo returns [] instead of set(), so handle that case directly
        return self.to_set_of_tuples(snapshot)

    def get_subscription(self, service_id):
        """Return publisher/tables subscribed by service: service_id

        Return data structure:
        {publisher_id -> set of tables}
        """
        result = {}
        for publisher in self.subscriptions:
            for table in self.subscriptions[publisher]:
                if service_id in self.subscriptions[publisher][table]:
                    try:
                        result[publisher].add(table)
                    except KeyError:
                        result[publisher] = set([table])
        return result

    def to_set_of_tuples(self, snapshot):
        try:
            return set([tuple(x) for x in snapshot])
        except TypeError:
            return snapshot

    def unsubscribe_table(self, subscriber, publisher, table):
        """Remove subscription for local service to target/table."""
        if publisher not in self.subscriptions:
            return False
        if table not in self.subscriptions[publisher]:
            return False
        self.subscriptions[publisher][table].discard(subscriber)
        if len(self.subscriptions[publisher][table]) == 0:
            del self.subscriptions[publisher][table]
        if len(self.subscriptions[publisher]) == 0:
            del self.subscriptions[publisher]

    def _update_tables_with_subscriber(self):
        # not thread-safe: assumes each dseNode is single-threaded
        peers = self.dse_status()['peers']
        for s in self.get_services():
            sid = s.service_id
            # first, include subscriptions within the node, if any
            tables_with_subs = set(self.subscriptions.get(sid, {}))
            # then add subscriptions from other nodes
            for peer_id in peers:
                if sid in peers[peer_id]['subscribed_tables']:
                    tables_with_subs |= peers[
                        peer_id]['subscribed_tables'][sid]
            s._published_tables_with_subscriber = tables_with_subs

    # Driver CRUD.  Maybe belongs in a subclass of DseNode?

    def load_drivers(self):
        """Load all configured drivers and check no name conflict"""
        result = {}
        for driver_path in cfg.CONF.drivers:
            obj = importutils.import_class(driver_path)
            driver = obj.get_datasource_info()
            if driver['id'] in result:
                raise exception.BadConfig(_("There is a driver loaded already"
                                          "with the driver name of %s")
                                          % driver['id'])
            driver['module'] = driver_path
            result[driver['id']] = driver
        return result

    def get_driver_info(self, driver):
        driver = self.loaded_drivers.get(driver)
        if not driver:
            raise exception.DriverNotFound(id=driver)
        return driver

    # Datasource CRUD.  Maybe belongs in a subclass of DseNode?

    def get_datasource(cls, id_):
        """Return the created datasource."""
        result = datasources_db.get_datasource(id_)
        if not result:
            raise exception.DatasourceNotFound(id=id_)
        return cls.make_datasource_dict(result)

    def get_datasources(self, filter_secret=False):
        """Return the created datasources as recorded in the DB.

        This returns what datasources the database contains, not the
        datasources that this server instance is running.
        """
        results = []
        for datasource in datasources_db.get_datasources():
            result = self.make_datasource_dict(datasource)
            if filter_secret:
                # driver_info knows which fields should be secret
                driver_info = self.get_driver_info(result['driver'])
                try:
                    for hide_field in driver_info['secret']:
                        result['config'][hide_field] = "<hidden>"
                except KeyError:
                    pass
            results.append(result)
        return results

    def make_datasource_dict(self, req, fields=None):
        result = {'id': req.get('id') or uuidutils.generate_uuid(),
                  'name': req.get('name'),
                  'driver': req.get('driver'),
                  'description': req.get('description'),
                  'type': None,
                  'enabled': req.get('enabled', True)}
        # NOTE(arosen): we store the config as a string in the db so
        # here we serialize it back when returning it.
        if isinstance(req.get('config'), six.string_types):
            result['config'] = json.loads(req['config'])
        else:
            result['config'] = req.get('config')

        return self._fields(result, fields)

    def _fields(self, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    def add_datasource(self, item, deleted=False, update_db=True):
        req = self.make_datasource_dict(item)
        # If update_db is True, new_id will get a new value from the db.
        new_id = req['id']
        driver_info = self.get_driver_info(item['driver'])
        session = db.get_session()
        try:
            with session.begin(subtransactions=True):
                LOG.debug("adding datasource %s", req['name'])
                if update_db:
                    LOG.debug("updating db")
                    datasource = datasources_db.add_datasource(
                        id_=req['id'],
                        name=req['name'],
                        driver=req['driver'],
                        config=req['config'],
                        description=req['description'],
                        enabled=req['enabled'],
                        session=session)
                    new_id = datasource['id']

                self.validate_create_datasource(req)
                if self.is_valid_service(req['name']):
                    raise exception.DatasourceNameInUse(value=req['name'])
                try:
                    self.create_service(
                        class_path=driver_info['module'],
                        kwargs={'name': req['name'], 'args': item['config']})
                except Exception:
                    raise exception.DatasourceCreationError(value=req['name'])

        except db_exc.DBDuplicateEntry:
            raise exception.DatasourceNameInUse(value=req['name'])
        new_item = dict(item)
        new_item['id'] = new_id
        return self.make_datasource_dict(new_item)

    def validate_create_datasource(self, req):
        driver = req['driver']
        config = req['config'] or {}
        for loaded_driver in self.loaded_drivers.values():
            if loaded_driver['id'] == driver:
                specified_options = set(config.keys())
                valid_options = set(loaded_driver['config'].keys())
                # Check that all the specified options passed in are
                # valid configuration options that the driver exposes.
                invalid_options = specified_options - valid_options
                if invalid_options:
                    raise exception.InvalidDriverOption(
                        invalid_options=invalid_options)

                # check that all the required options are passed in
                required_options = set(
                    [k for k, v in loaded_driver['config'].items()
                     if v == constants.REQUIRED])
                missing_options = required_options - specified_options
                if missing_options:
                    missing_options = ', '.join(missing_options)
                    raise exception.MissingRequiredConfigOptions(
                        missing_options=missing_options)
                return loaded_driver

        # If we get here no datasource driver match was found.
        raise exception.InvalidDriver(driver=req)

    def create_service(self, class_path, kwargs):
        """Create a new DataService on this node.

        :param name is the name of the service.  Must be unique across all
               services
        :param classPath is a string giving the path to the class name, e.g.
               congress.datasources.fake_datasource.FakeDataSource
        :param args is the list of arguments to give the DataService
               constructor
        :param type_ is the kind of service
        :param id_ is an optional parameter for specifying the uuid.
        """

        # TODO(dse2): fix logging.  Want to show kwargs, but hide passwords.
        # self.log_info("creating service %s with class %s and args %s",
        #               name, moduleName, strutils.mask_password(args, "****"))

        # split class_path into module and class name
        pieces = class_path.split(".")
        module_name = ".".join(pieces[:-1])
        class_name = pieces[-1]

        # import the module
        try:
            module = importutils.import_module(module_name)
            service = getattr(module, class_name)(**kwargs)
            self.register_service(service)
        except Exception:
            # TODO(dse2): add logging for service creation failure
            raise exception.DataServiceError(
                "Error loading instance of module '%s':: \n%s"
                % (class_path, traceback.format_exc()))

    def delete_datasource(self, datasource, update_db=True):
        datasource_id = datasource['id']
        session = db.get_session()
        with session.begin(subtransactions=True):
            if update_db:
                result = datasources_db.delete_datasource(
                    datasource_id, session)
                if not result:
                    raise exception.DatasourceNotFound(id=datasource_id)
            self.unregister_service(datasource['name'])


class DseNodeEndpoints (object):
    """Collection of RPC endpoints that the DseNode exposes on the bus.

       Must be a separate class since all public methods of a given
       class are assumed to be valid RPC endpoints.
    """

    def __init__(self, dsenode):
        self.node = dsenode

    def handle_publish(self, context, publisher, table, data):
        """Function called on the node when a publication is sent.

           Forwards the publication to all of the relevant services.
        """
        for s in self.node.table_subscribers(publisher, table):
            self.node.service_object(s).receive_data(
                publisher=publisher, table=table, data=data)



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import base
from congress.datalog import compile
from congress.datalog import ruleset
from congress.datalog import topdown
from congress.datalog import utility
from congress import exception


LOG = logging.getLogger(__name__)


class NonrecursiveRuleTheory(topdown.TopDownTheory):
    """A non-recursive collection of Rules."""

    def __init__(self, name=None, abbr=None,
                 schema=None, theories=None, desc=None, owner=None):
        super(NonrecursiveRuleTheory, self).__init__(
            name=name, abbr=abbr, theories=theories, schema=schema,
            desc=desc, owner=owner)
        # dictionary from table name to list of rules with that table in head
        self.rules = ruleset.RuleSet()
        self.kind = base.NONRECURSIVE_POLICY_TYPE
        if schema is None:
            self.schema = compile.Schema()

    # External Interface

    # SELECT implemented by TopDownTheory

    def initialize_tables(self, tablenames, facts):
        """Event handler for (re)initializing a collection of tables

        @facts must be an iterable containing compile.Fact objects.
        """
        LOG.info("initialize_tables")
        cleared_tables = set(tablenames)
        for t in tablenames:
            self.rules.clear_table(t)

        count = 0
        extra_tables = set()
        ignored_facts = 0
        for f in facts:
            if f.table not in cleared_tables:
                extra_tables.add(f.table)
                ignored_facts += 1
            else:
                self.rules.add_rule(f.table, f)
                count += 1
                if self.schema:
                    self.schema.update(f, True)
        if ignored_facts > 0:
            LOG.error("initialize_tables ignored %d facts for tables "
                      "%s not included in the list of tablenames %s",
                      ignored_facts, extra_tables, cleared_tables)
        LOG.info("initialized %d tables with %d facts",
                 len(cleared_tables), count)

    def insert(self, rule):
        changes = self.update([compile.Event(formula=rule, insert=True)])
        return [event.formula for event in changes]

    def delete(self, rule):
        changes = self.update([compile.Event(formula=rule, insert=False)])
        return [event.formula for event in changes]

    def _update_lit_schema(self, lit, is_insert):
        if self.schema is None:
            raise exception.PolicyException(
                "Cannot update schema because theory %s doesn't have "
                "schema." % self.name)

        if self.schema.complete:
            # complete means the schema is pre-built and shouldn't be updated
            return None
        return self.schema.update(lit, is_insert)

    def update_rule_schema(self, rule, is_insert):
        schema_changes = []
        if self.schema is None or not self.theories or self.schema.complete:
            # complete means the schema is pre-built like datasoures'
            return schema_changes

        if isinstance(rule, compile.Fact) or isinstance(rule, compile.Literal):
            schema_changes.append(self._update_lit_schema(rule, is_insert))
            return schema_changes

        schema_changes.append(self._update_lit_schema(rule.head, is_insert))

        for lit in rule.body:
            if lit.is_builtin():
                continue
            active_theory = lit.table.service or self.name
            if active_theory not in self.theories:
                continue
            schema_changes.append(
                self.theories[active_theory]._update_lit_schema(lit,
                                                                is_insert))

        return schema_changes

    def revert_schema(self, schema_changes):
        if not self.theories:
            return
        for change in schema_changes:
            if not change:
                continue
            active_theory = change[3]
            if not active_theory:
                self.schema.revert(change)
            else:
                self.theories[active_theory].schema.revert(change)

    def update(self, events):
        """Apply EVENTS.

           And return the list of EVENTS that actually
           changed the theory.  Each event is the insert or delete of
           a policy statement.
           """
        changes = []
        self.log(None, "Update %s", utility.iterstr(events))
        try:
            for event in events:
                schema_changes = self.update_rule_schema(
                    event.formula, event.insert)
                formula = compile.reorder_for_safety(event.formula)
                if event.insert:
                    if self._insert_actual(formula):
                        changes.append(event)
                    else:
                        self.revert_schema(schema_changes)
                else:
                    if self._delete_actual(formula):
                        changes.append(event)
                    else:
                        self.revert_schema(schema_changes)
        except Exception as e:
            LOG.exception("runtime caught an exception")
            raise e

        return changes

    def update_would_cause_errors(self, events):
        """Return a list of PolicyException.

        Return a list of PolicyException if we were
        to apply the insert/deletes of policy statements dictated by
        EVENTS to the current policy.
        """
        self.log(None, "update_would_cause_errors %s", utility.iterstr(events))
        errors = []
        for event in events:
            if not compile.is_datalog(event.formula):
                errors.append(exception.PolicyException(
                    "Non-formula found: {}".format(
                        str(event.formula))))
            else:
                if event.formula.is_atom():
                    errors.extend(compile.fact_errors(
                        event.formula, self.theories, self.name))
                else:
                    errors.extend(compile.rule_errors(
                        event.formula, self.theories, self.name))
        # Would also check that rules are non-recursive, but that
        #   is currently being handled by Runtime.  The current implementation
        #   disallows recursion in all theories.
        return errors

    def define(self, rules):
        """Empties and then inserts RULES."""
        self.empty()
        return self.update([compile.Event(formula=rule, insert=True)
                            for rule in rules])

    def empty(self, tablenames=None, invert=False):
        """Deletes contents of theory.

        If provided, TABLENAMES causes only the removal of all rules
        that help define one of the tables in TABLENAMES.
        If INVERT is true, all rules defining anything other than a
        table in TABLENAMES is deleted.
        """
        if tablenames is None:
            self.rules.clear()
            return
        if invert:
            to_clear = set(self.defined_tablenames()) - set(tablenames)
        else:
            to_clear = tablenames
        for table in to_clear:
            self.rules.clear_table(table)

    def policy(self):
        # eliminate all rules with empty bodies
        return [p for p in self.content() if len(p.body) > 0]

    def __contains__(self, formula):
        if compile.is_atom(formula):
            return self.rules.contains(formula.table.table, formula)
        else:
            return self.rules.contains(formula.head.table.table, formula)

    # Internal Interface

    def _insert_actual(self, rule):
        """Insert RULE and return True if there was a change."""
        if compile.is_atom(rule):
            rule = compile.Rule(rule, [], rule.location)
        self.log(rule.head.table.table, "Insert: %s", repr(rule))
        return self.rules.add_rule(rule.head.table.table, rule)

    def _delete_actual(self, rule):
        """Delete RULE and return True if there was a change."""
        if compile.is_atom(rule):
            rule = compile.Rule(rule, [], rule.location)
        self.log(rule.head.table.table, "Delete: %s", rule)
        return self.rules.discard_rule(rule.head.table.table, rule)

    def content(self, tablenames=None):
        if tablenames is None:
            tablenames = self.rules.keys()
        results = []
        for table in tablenames:
            if table in self.rules:
                results.extend(self.rules.get_rules(table))
        return results

    def head_index(self, table, match_literal=None):
        """Return head index.

        This routine must return all the formulas pertinent for
        top-down evaluation when a literal with TABLE is at the top
        of the stack.
        """
        if table in self.rules:
            return self.rules.get_rules(table, match_literal)
        return []

    def arity(self, table, modal=None):
        """Return the number of arguments TABLENAME takes.

        :param table can be either a string or a Tablename
        Returns None if arity is unknown (if it does not occur in
            the head of a rule).
        """
        if isinstance(table, compile.Tablename):
            service = table.service
            name = table.table
            fullname = table.name()
        else:
            fullname = table
            service, name = compile.Tablename.parse_service_table(table)
        # check if schema knows the answer
        if self.schema:
            if service is None or service == self.name:
                arity = self.schema.arity(name)
            else:
                arity = self.schema.arity(fullname)
            if arity is not None:
                return arity
        # assuming a single arity for all tables
        formulas = self.head_index(fullname) or self.head_index(name)
        try:
            first = next(f for f in formulas
                         if f.head.table.matches(service, name, modal))
        except StopIteration:
            return None
        # should probably have an overridable function for computing
        #   the arguments of a head.  Instead we assume heads have .arguments
        return len(self.head(first).arguments)

    def defined_tablenames(self):
        """Returns list of table names defined in/written to this theory."""
        return self.rules.keys()

    def head(self, formula):
        """Given the output from head_index(), return the formula head.

        Given a FORMULA, return the thing to unify against.
        Usually, FORMULA is a compile.Rule, but it could be anything
        returned by HEAD_INDEX.
        """
        return formula.head

    def body(self, formula):
        """Return formula body.

        Given a FORMULA, return a list of things to push onto the
        top-down eval stack.
        """
        return formula.body


class ActionTheory(NonrecursiveRuleTheory):
    """ActionTheory object.

    Same as NonrecursiveRuleTheory except it has fewer constraints
    on the permitted rules. Still working out the details.
    """
    def __init__(self, name=None, abbr=None,
                 schema=None, theories=None, desc=None, owner=None):
        super(ActionTheory, self).__init__(name=name, abbr=abbr,
                                           schema=schema, theories=theories,
                                           desc=desc, owner=owner)
        self.kind = base.ACTION_POLICY_TYPE

    def update_would_cause_errors(self, events):
        """Return a list of PolicyException.

        Return a list of PolicyException if we were
        to apply the events EVENTS to the current policy.
        """
        self.log(None, "update_would_cause_errors %s", utility.iterstr(events))
        errors = []
        for event in events:
            if not compile.is_datalog(event.formula):
                errors.append(exception.PolicyException(
                    "Non-formula found: {}".format(
                        str(event.formula))))
            else:
                if event.formula.is_atom():
                    errors.extend(compile.fact_errors(
                        event.formula, self.theories, self.name))
                else:
                    errors.extend(compile.rule_head_has_no_theory(
                        event.formula,
                        permit_head=lambda lit: lit.is_update()))
                    # Should put this back in place, but there are some
                    # exceptions that we don't handle right now.
                    # Would like to mark some tables as only being defined
                    #   for certain bound/free arguments and take that into
                    #   account when doing error checking.
                    # errors.extend(compile.rule_negation_safety(event.formula))
        return errors


class MultiModuleNonrecursiveRuleTheory(NonrecursiveRuleTheory):
    """MultiModuleNonrecursiveRuleTheory object.

    Same as NonrecursiveRuleTheory, except we allow rules with theories
    in the head.  Intended for use with TopDownTheory's INSTANCES method.
    """
    def _insert_actual(self, rule):
        """Insert RULE and return True if there was a change."""
        if compile.is_atom(rule):
            rule = compile.Rule(rule, [], rule.location)
        self.log(rule.head.table.table, "Insert: %s", rule)
        return self.rules.add_rule(rule.head.table.table, rule)

    def _delete_actual(self, rule):
        """Delete RULE and return True if there was a change."""
        if compile.is_atom(rule):
            rule = compile.Rule(rule, [], rule.location)
        self.log(rule.head.table.table, "Delete: %s", rule)
        return self.rules.discard_rule(rule.head.table.table, rule)

    # def update_would_cause_errors(self, events):
    #     return []


class DatasourcePolicyTheory(NonrecursiveRuleTheory):
    """DatasourcePolicyTheory

    DatasourcePolicyTheory is identical to NonrecursiveRuleTheory, except that
    self.kind is base.DATASOURCE_POLICY_TYPE instead of
    base.NONRECURSIVE_POLICY_TYPE.  DatasourcePolicyTheory uses a different
    self.kind so that the synchronizer knows not to synchronize policies of
    kind DatasourcePolicyTheory with the database listing of policies.
"""

    def __init__(self, name=None, abbr=None,
                 schema=None, theories=None, desc=None, owner=None):
        super(DatasourcePolicyTheory, self).__init__(
            name=name, abbr=abbr, theories=theories, schema=schema,
            desc=desc, owner=owner)
        self.kind = base.DATASOURCE_POLICY_TYPE



#! /usr/bin/python
#
# Copyright (c) 2014 IBM, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from oslo_log import log as logging

from congress.datalog import base as datalog_base
from congress.datalog.builtin import congressbuiltin
from congress.datalog import compile
from congress import exception
from congress.policy_engines import agnostic
from congress.tests import base
from congress.tests import helper

LOG = logging.getLogger(__name__)

addmap = {
    'comparison': [
        {'func': 'f(x,y)', 'num_inputs': 2,
         'code': lambda x, y: x if x > y else y}],
    'newcategory': [
        {'func': 'g(x,y)', 'num_inputs': 2, 'code': lambda x, y: x + y}]}


append_builtin = {'arithmetic': [{'func': 'div(x,y)',
                                  'num_inputs': 2,
                                  'code': 'lambda x,y: x / y'}]}


class TestBuiltins(base.TestCase):
    def setUp(self):
        super(TestBuiltins, self).setUp()
        self.cbcmap = congressbuiltin.CongressBuiltinCategoryMap(
            congressbuiltin._builtin_map)
        self.predl = self.cbcmap.builtin('lt')

    def test_add_and_delete_map(self):
        cbcmap_before = self.cbcmap
        self.cbcmap.add_map(append_builtin)
        self.cbcmap.delete_map(append_builtin)
        self.assertTrue(self.cbcmap.mapequal(cbcmap_before))

    def test_add_map_only(self):
        self.cbcmap.add_map(append_builtin)
        predl = self.cbcmap.builtin('div')
        self.assertNotEqual(predl, None)
        self.cbcmap.add_map(addmap)
        predl = self.cbcmap.builtin('max')
        self.assertNotEqual(predl, None)

    def test_add_and_delete_builtin(self):
        cbcmap_before = self.cbcmap
        self.cbcmap.add_map(append_builtin)
        self.cbcmap.delete_builtin('arithmetic', 'div', 2)
        self.assertTrue(self.cbcmap.mapequal(cbcmap_before))

    def test_string_pred_string(self):
        predstring = str(self.predl)
        self.assertNotEqual(predstring, 'ltc(x,y')

    def test_add_and_delete_to_category(self):
        cbcmap_before = self.cbcmap
        arglist = ['x', 'y', 'z']
        pred = congressbuiltin.CongressBuiltinPred('testfunc', arglist, 1,
                                                   lambda x: not x)
        self.cbcmap.insert_to_category('arithmetic', pred)
        self.cbcmap.delete_from_category('arithmetic', pred)
        self.assertTrue(self.cbcmap.mapequal(cbcmap_before))

    def test_all_checks(self):
        predtotest = self.cbcmap.builtin('lt')
        self.assertTrue(self.cbcmap.builtin_is_registered(predtotest))

    def test_eval_builtin(self):
        predl = self.cbcmap.builtin('plus')
        result = predl.code(1, 2)
        self.assertEqual(result, 3)
        predl = self.cbcmap.builtin('gt')
        result = predl.code(1, 2)
        self.assertEqual(result, False)


class TestReorder(base.TestCase):
    def check(self, input_string, correct_string, msg):
        rule = compile.parse1(input_string)
        actual = compile.reorder_for_safety(rule)
        correct = compile.parse1(correct_string)
        if correct != actual:
            emsg = "Correct: " + str(correct)
            emsg += "; Actual: " + str(actual)
            self.fail(msg + " :: " + emsg)

    def check_err(self, input_string, unsafe_lit_strings, msg):
        rule = compile.parse1(input_string)
        try:
            compile.reorder_for_safety(rule)
            self.fail("Failed to raise exception for " + input_string)
        except exception.PolicyException as e:
            errmsg = str(e)
            # parse then print to string so string rep same in err msg
            unsafe_lits = [str(compile.parse1(x)) for x in unsafe_lit_strings]
            missing_lits = [m for m in unsafe_lits
                            if m + " (vars" not in errmsg]
            if len(missing_lits) > 0:
                self.fail(
                    "Unsafe literals {} not reported in error: {}".format(
                        ";".join(missing_lits), errmsg))

    def test_reorder_builtins(self):
        self.check("p(x, z) :- q(x, y), plus(x, y, z)",
                   "p(x, z) :- q(x, y), plus(x, y, z)",
                   "No reorder")

        self.check("p(x, z) :- plus(x, y, z), q(x, y)",
                   "p(x, z) :- q(x, y), plus(x, y, z)",
                   "Basic reorder")

        self.check("p(x, z) :- q(x, y), r(w), plus(x, y, z), plus(z, w, y)",
                   "p(x, z) :- q(x, y), r(w), plus(x, y, z), plus(z, w, y)",
                   "Chaining: no reorder")

        self.check("p(x, z) :- q(x, y),  plus(x, y, z), plus(z, w, y), r(w)",
                   "p(x, z) :- q(x, y), plus(x, y, z), r(w), plus(z, w, y)",
                   "Chaining: reorder")

        self.check("p(x) :- lt(t, v), plus(z, w, t), plus(z, u, v), "
                   "   plus(x, y, z), q(y), r(x), s(u), t(w) ",
                   "p(x) :- q(y), r(x), plus(x, y, z), s(u), plus(z, u, v), "
                   "   t(w), plus(z, w, t), lt(t, v)",
                   "Partial-order chaining")

    def test_unsafe_builtins(self):
        # an output
        self.check_err("p(x) :- q(x), plus(x, y, z)",
                       ["plus(x,y,z)"],
                       "Basic Unsafe input")

        self.check_err("p(x) :- q(x), r(z), plus(x, y, z)",
                       ["plus(x,y,z)"],
                       "Basic Unsafe input 2")

        self.check_err("p(x, z) :- plus(x, y, z), plus(z, y, x), "
                       "   plus(x, z, y)",
                       ["plus(x, y, z)", "plus(z, y, x)", "plus(x, z, y)"],
                       "Unsafe with cycle")

        # no outputs
        self.check_err("p(x) :- q(x), lt(x, y)",
                       ["lt(x,y)"],
                       "Basic Unsafe input, no outputs")

        self.check_err("p(x) :- q(y), lt(x, y)",
                       ["lt(x,y)"],
                       "Basic Unsafe input, no outputs 2")

        self.check_err("p(x, z) :- lt(x, y), lt(y, x)",
                       ["lt(x,y)", "lt(y, x)"],
                       "Unsafe with cycle, no outputs")

        # chaining
        self.check_err("p(x) :- q(x, y), plus(x, y, z), plus(z, 3, w), "
                       "    plus(w, t, u)",
                       ["plus(w, t, u)"],
                       "Unsafe chaining")

        self.check_err("p(x) :- q(x, y), plus(x, y, z), plus(z, 3, w), "
                       "   lt(w, t)",
                       ["lt(w, t)"],
                       "Unsafe chaining 2")

    def test_reorder_negation(self):
        self.check("p(x) :- q(x), not u(x), r(y), not s(x, y)",
                   "p(x) :- q(x), not u(x), r(y), not s(x, y)",
                   "No reordering")

        self.check("p(x) :- not q(x), r(x)",
                   "p(x) :- r(x), not q(x)",
                   "Basic")

        self.check("p(x) :- r(x), not q(x, y), s(y)",
                   "p(x) :- r(x), s(y), not q(x,y)",
                   "Partially safe")

        self.check("p(x) :- not q(x, y), not r(x), not r(x, z), "
                   "    t(x, y), u(x), s(z)",
                   "p(x) :- t(x,y), not q(x,y), not r(x), u(x), s(z), "
                   "    not r(x, z)",
                   "Complex")

    def test_unsafe_negation(self):
        self.check_err("p(x) :- not q(x)",
                       ["q(x)"],
                       "Basic")

        self.check_err("p(x) :- not q(x), not r(x)",
                       ["q(x)", "r(x)"],
                       "Cycle")

        self.check_err("p(x) :- not q(x, y), r(y)",
                       ["q(x, y)"],
                       "Partially safe")

    def test_reorder_builtins_negation(self):
        self.check("p(x) :- not q(z), plus(x, y, z), s(x), s(y)",
                   "p(x) :- s(x), s(y), plus(x, y, z), not q(z)",
                   "Basic")

        self.check("p(x) :- not q(z, w), plus(x, y, z), lt(z, w), "
                   "   plus(x, 3, w), s(x, y)",
                   "p(x) :- s(x,y), plus(x, y, z), plus(x, 3, w), "
                   "   not q(z, w), lt(z, w)",
                   "Partial order")

    def test_unsafe_builtins_negation(self):
        self.check_err("p(x) :- plus(x, y, z), not q(x, y)",
                       ['plus(x,y,z)', 'q(x,y)'],
                       'Unsafe cycle')

        self.check_err("p(x) :- plus(x, y, z), plus(z, w, t), not q(z, t),"
                       "    s(x), t(y)",
                       ['plus(z, w, t)', 'q(z, t)'],
                       'Unsafety propagates')


NREC_THEORY = 'non-recursive theory test'
MAT_THEORY = 'materialized view theory test'


class TestTheories(base.TestCase):
    def prep_runtime(self, code=None, msg=None, target=None):
        # compile source
        if msg is not None:
            LOG.debug(msg)
        if code is None:
            code = ""
        if target is None:
            target = NREC_THEORY
        run = agnostic.Runtime()
        run.create_policy(NREC_THEORY, abbr="NRT",
                          kind=datalog_base.NONRECURSIVE_POLICY_TYPE)
        run.create_policy(MAT_THEORY, abbr="MAT",
                          kind=datalog_base.MATERIALIZED_POLICY_TYPE)
        run.debug_mode()
        run.insert(code, target=target)
        return run

    def check_equal(self, actual_string, correct_string, msg):
        self.assertTrue(helper.datalog_equal(
            actual_string, correct_string, msg))

    def test_materialized_builtins(self):
        self.test_builtins(MAT_THEORY)

    def test_builtins(self, th=NREC_THEORY):
        """Test the mechanism that implements builtins."""
        run = self.prep_runtime()
        run.insert('p(x) :- q(x,y), plus(x,y,z), r(z)'
                   'q(1,2)'
                   'q(2,3)'
                   'r(3)'
                   'r(5)', target=th)
        self.check_equal(run.select('p(x)', target=th), "p(1) p(2)", "Plus")
        run.delete('r(5)', target=th)
        self.check_equal(run.select('p(x)', target=th), "p(1)", "Plus")

        run = self.prep_runtime()
        run.insert('p(x) :- q(x,y), minus(x,y,z), r(z)'
                   'q(2,1)'
                   'q(3,1)'
                   'r(1)'
                   'r(4)', target=th)
        self.check_equal(run.select('p(x)', target=th), "p(2)", "Minus")
        run.delete('r(4)', target=th)
        run.insert('r(2)', target=th)
        self.check_equal(run.select('p(x)', target=th), "p(2) p(3)", "Minus")

        run = self.prep_runtime()
        run.insert('p(x, z) :- q(x,y), plus(x,y,z)'
                   'q(1,2)'
                   'q(2,3)', target=th)
        self.check_equal(run.select('p(x, y)', target=th),
                         "p(1, 3) p(2, 5)", "Plus")

        run = self.prep_runtime()
        run.insert('m(x) :- j(x,y), lt(x,y)'
                   'j(1,2)'
                   'j(3,2)', target=th)
        self.check_equal(run.select('m(x)', target=th), 'm(1)', "LT")

        run = self.prep_runtime()
        run.insert('m(x) :- j(x,y), lt(x,y), r(y)'
                   'j(1,2)'
                   'j(2,3)'
                   'j(3,2)'
                   'r(2)', target=th)
        self.check_equal(run.select('m(x)', target=th), 'm(1)', "LT 2")

        run = self.prep_runtime()
        run.insert('p(x,z) :- q(x), plus(x,1,z)'
                   'q(3)'
                   'q(5)', target=th)
        self.check_equal(run.select('p(x,z)', target=th),
                         'p(3, 4) p(5,6)', "Bound input")

        run = self.prep_runtime()
        run.insert('p(x) :- q(x), plus(x,1,5)'
                   'q(4)'
                   'q(5)', target=th)
        self.check_equal(run.select('p(x)', target=th),
                         'p(4)', "Bound output")

        run = self.prep_runtime()
        run.insert('p(x, z) :- plus(x,y,z), q(x), r(y)'
                   'q(4)'
                   'r(5)', target=th)
        self.check_equal(run.select('p(x, y)', target=th),
                         'p(4, 9)',
                         "Reordering")

        run = self.prep_runtime()
        run.insert('p(x, z) :- plus(x,y,z), q(x), q(y)'
                   'q(4)'
                   'q(5)', target=th)
        self.check_equal(run.select('p(x, y)', target=th),
                         'p(4, 9) p(4, 8) p(5, 9) p(5, 10)',
                         "Reordering with self joins")

    def test_materialized_builtins_content(self):
        self.test_builtins_content(MAT_THEORY)

    def test_builtins_content(self, th=NREC_THEORY):
        """Test the content of the builtins, not the mechanism."""
        def check_true(code, msg):
            run = self.prep_runtime('')
            run.insert(code, target=th)
            self.check_equal(
                run.select('p(x)', target=th),
                'p(1)',
                msg)

        def check_false(code, msg):
            th = NREC_THEORY
            run = self.prep_runtime('')
            run.insert(code, target=th)
            self.check_equal(
                run.select('p(x)', target=th),
                '',
                msg)

        #
        # Numbers
        #

        # int
        code = 'p(1) :- int(2,2)'
        check_true(code, "int")

        code = 'p(1) :- int(2.3, 2)'
        check_true(code, "int")

        code = 'p(1) :- int(2, 3.3)'
        check_false(code, "int")

        # float
        code = 'p(1) :- float(2,2.0)'
        check_true(code, "float")

        code = 'p(1) :- float(2.3,2.3)'
        check_true(code, "float")

        code = 'p(1) :- float(2,3.3)'
        check_false(code, "int")

        # plus
        code = 'p(1) :- plus(2,3,5)'
        check_true(code, "plus")

        code = 'p(1) :- plus(2,3,1)'
        check_false(code, "plus")

        # minus
        code = 'p(1) :- minus(5, 3, 2)'
        check_true(code, "minus")

        code = 'p(1) :- minus(5, 3, 6)'
        check_false(code, "minus")

        # minus negative: negative numbers should not be supported
        # code = 'p(1) :- minus(3, 5, x)'
        # check_false(code, "minus")

        # times
        code = 'p(1) :- mul(3, 5, 15)'
        check_true(code, "multiply")

        code = 'p(1) :- mul(2, 5, 1)'
        check_false(code, "multiply")

        # divides
        code = 'p(1) :- div(10, 2, 5)'
        check_true(code, "divides")

        code = 'p(1) :- div(10, 4, 2)'
        check_true(code, "integer divides")

        code = 'p(1) :- div(10, 4.0, 2.5)'
        check_true(code, "float divides")

        code = 'p(1) :- div(10.0, 3, 3.3)'
        check_false(code, "divides")

        #
        # Comparison
        #

        # less than
        code = 'p(1) :- lt(1, 3)'
        check_true(code, "lessthan")

        code = 'p(1) :- lt(5, 2)'
        check_false(code, "lessthan")

        # less than equal
        code = 'p(1) :- lteq(1, 3)'
        check_true(code, "lessthaneq")

        code = 'p(1) :- lteq(3, 3)'
        check_true(code, "lessthaneq")

        code = 'p(1) :- lteq(4, 3)'
        check_false(code, "lessthaneq")

        # greater than
        code = 'p(1) :- gt(9, 5)'
        check_true(code, "greaterthan")

        code = 'p(1) :- gt(5, 9)'
        check_false(code, "greaterthan")

        # greater than equal
        code = 'p(1) :- gteq(10, 5)'
        check_true(code, "greaterthaneq")

        code = 'p(1) :- gteq(10, 10)'
        check_true(code, "greaterthaneq")

        code = 'p(1) :- gteq(5, 20)'
        check_false(code, "greaterthaneq")

        # equal
        code = 'p(1) :- equal(5, 5)'
        check_true(code, "equal")

        code = 'p(1) :- equal(5, 7)'
        check_false(code, "equal")

        # max
        code = 'p(1) :- max(3, 4, 4)'
        check_true(code, "max")

        code = 'p(1) :- max(3, 7, 3)'
        check_false(code, "max")

        #
        # Strings
        #

        # len
        code = 'p(1) :- len("abcde", 5)'
        check_true(code, "Len")

        code = 'p(1) :- len("abcde", 7)'
        check_false(code, "Len")

        # concat
        code = 'p(1) :- concat("abc", "def", "abcdef")'
        check_true(code, "concat")

        code = 'p(1) :- concat("abc", "def", "zxy")'
        check_false(code, "concat")

        #
        # Datetime
        # We should make some of these more robust but can't do
        #   that with the safety restrictions in place at the time
        #   of writing.
        #

        # lessthan
        code = ('p(1) :- datetime_lt('
                '"Jan 1, 2014 10:00:00", "2014-01-02 10:00:00")')
        check_true(code, "True datetime_lt")

        code = ('p(1) :- datetime_lt('
                '"2014-01-03 10:00:00", "Jan 2, 2014 10:00:00")')
        check_false(code, "False datetime_lt")

        # lessthanequal
        code = ('p(1) :- datetime_lteq('
                '"Jan 1, 2014 10:00:00", "2014-01-02 10:00:00")')
        check_true(code, "True datetime_lteq")

        code = ('p(1) :- datetime_lteq('
                '"Jan 1, 2014 10:00:00", "2014-01-01 10:00:00")')
        check_true(code, "True datetime_lteq")

        code = ('p(1) :- datetime_lteq('
                '"2014-01-02 10:00:00", "Jan 1, 2014 10:00:00")')
        check_false(code, "False datetime_lteq")

        # greaterthan
        code = ('p(1) :- datetime_gt('
                '"Jan 5, 2014 10:00:00", "2014-01-02 10:00:00")')
        check_true(code, "True datetime_gt")

        code = ('p(1) :- datetime_gt('
                '"2014-01-03 10:00:00", "Feb 2, 2014 10:00:00")')
        check_false(code, "False datetime_gt")

        # greaterthanequal
        code = ('p(1) :- datetime_gteq('
                '"Jan 5, 2014 10:00:00", "2014-01-02 10:00:00")')
        check_true(code, "True datetime_gteq")

        code = ('p(1) :- datetime_gteq('
                '"Jan 5, 2014 10:00:00", "2014-01-05 10:00:00")')
        check_true(code, "True datetime_gteq")

        code = ('p(1) :- datetime_gteq('
                '"2014-01-02 10:00:00", "Mar 1, 2014 10:00:00")')
        check_false(code, "False datetime_gteq")

        # equal
        code = ('p(1) :- datetime_equal('
                '"Jan 5, 2014 10:00:00", "2014-01-05 10:00:00")')
        check_true(code, "True datetime_equal")

        code = ('p(1) :- datetime_equal('
                '"Jan 5, 2014 10:00:00", "2014-01-02 10:00:00")')
        check_false(code, "False datetime_equal")

        # plus
        code = ('p(1) :- datetime_plus('
                '"Jan 5, 2014 10:00:00", 3600, "2014-01-05 11:00:00")')
        check_true(code, "True datetime_plus")

        code = ('p(1) :- datetime_plus('
                '"Jan 5, 2014 10:00:00", "1:00:00", "2014-01-05 11:00:00")')
        check_true(code, "True datetime_plus")

        code = ('p(1) :- datetime_plus('
                '"Jan 5, 2014 10:00:00", 3600, "2014-01-05 12:00:00")')
        check_false(code, "False datetime_plus")

        # minus
        code = ('p(1) :- datetime_minus('
                '"Jan 5, 2014 10:00:00", "25:00:00", "2014-01-04 09:00:00")')
        check_true(code, "True datetime_minus")

        code = ('p(1) :- datetime_minus('
                '"Jan 5, 2014 10:00:00", 3600, "2014-01-05 09:00:00")')
        check_true(code, "True datetime_minus")

        code = ('p(1) :- datetime_minus('
                '"Jan 5, 2014 10:00:00", "9:00:00", "Jan 4, 2014 10:00:00")')
        check_false(code, "False datetime_minus")

        # to_seconds
        code = ('p(1) :- datetime_to_seconds('
                '"Jan 1, 1900 1:00:00", 3600)')
        check_true(code, "True datetime_to_seconds")

        code = ('p(1) :- datetime_to_seconds('
                '"Jan 1, 1900 1:00:00", 3601)')
        check_false(code, "False datetime_to_seconds")

        # extract_time
        code = ('p(1) :- extract_time('
                '"Jan 1, 1900 1:00:00", "01:00:00")')
        check_true(code, "True extract_time")

        code = ('p(1) :- extract_time('
                '"Jan 1, 1900 1:00:00", "02:00:00")')
        check_false(code, "False extract_time")

        # extract_date
        code = ('p(1) :- extract_date('
                '"Jan 1, 1900 1:00:00", "1900-01-01")')
        check_true(code, "True extract_date")

        code = ('p(1) :- extract_date('
                '"Jan 1, 1900 1:00:00", "2000-01-01")')
        check_false(code, "False extract_date")

        # pack_datetime
        code = ('p(1) :- pack_datetime(2000, 1, 1, 10, 5, 6, '
                '"2000-1-1 10:5:6")')
        check_true(code, "True pack_datetime")

        code = ('p(1) :- pack_datetime(2000, 1, 1, 10, 5, 6, '
                '"2000-1-1 10:5:20")')
        check_false(code, "False pack_datetime")

        # pack_date
        code = ('p(1) :- pack_date(2000, 1, 1, '
                '"2000-1-1")')
        check_true(code, "True pack_date")

        code = ('p(1) :- pack_date(2000, 1, 1, '
                '"2000-1-2")')
        check_false(code, "False pack_date")

        # pack_time
        code = ('p(1) :- pack_time(5, 6, 7, '
                '"5:6:7")')
        check_true(code, "True pack_time")

        code = ('p(1) :- pack_time(5, 6, 7, '
                '"10:6:7")')
        check_false(code, "False pack_time")

        # unpack_datetime
        code = ('p(1) :- unpack_datetime("2000-1-1 10:5:6", '
                '2000, 1, 1, 10, 5, 6)')
        check_true(code, "True unpack_datetime")

        code = ('p(1) :- unpack_datetime("2000-1-1 10:5:6", '
                '2000, 1, 1, 12, 5, 6)')
        check_false(code, "False unpack_datetime")

        # unpack_date
        code = ('p(1) :- unpack_date("2000-1-1 10:5:6", '
                '2000, 1, 1)')
        check_true(code, "True unpack_date")

        code = ('p(1) :- unpack_date("2000-1-1 10:5:6", '
                '2000, 1, 5)')
        check_false(code, "False unpack_date")

        # unpack_time
        code = ('p(1) :- unpack_time("2000-1-1 10:5:6", '
                '10, 5, 6)')
        check_true(code, "True unpack_time")

        code = ('p(1) :- unpack_time("2000-1-1 10:5:6", '
                '12, 5, 6)')
        check_false(code, "False unpack_time")

        # unpack_time
        code = 'p(1) :- now(x)'
        check_true(code, "True unpack_time")



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import versions
from congress.api import webservice


class APIRouterV1(object):

    def __init__(self, resource_mgr, process_dict):
        """Bootstrap data models and handlers for the API definition."""

        # Setup /v1/
        version_v1_handler = versions.VersionV1Handler(r'/v1[/]?')
        resource_mgr.register_handler(version_v1_handler)

        policies = process_dict['api-policy']

        policy_collection_handler = webservice.CollectionHandler(
            r'/v1/policies',
            policies)
        resource_mgr.register_handler(policy_collection_handler)
        policy_path = r'/v1/policies/(?P<policy_id>[^/]+)'
        policy_element_handler = webservice.ElementHandler(
            policy_path,
            policies,
            policy_collection_handler,
            allow_update=False,
            allow_replace=False)
        resource_mgr.register_handler(policy_element_handler)

        policy_rules = process_dict['api-rule']
        rule_collection_handler = webservice.CollectionHandler(
            r'/v1/policies/(?P<policy_id>[^/]+)/rules',
            policy_rules,
            "{policy_id}")
        resource_mgr.register_handler(rule_collection_handler)
        rule_path = (r'/v1/policies/(?P<policy_id>[^/]+)' +
                     r'/rules/(?P<rule_id>[^/]+)')
        rule_element_handler = webservice.ElementHandler(
            rule_path,
            policy_rules,
            "{policy_id}")
        resource_mgr.register_handler(rule_element_handler)

        # Setup /v1/data-sources
        data_sources = process_dict['api-datasource']
        ds_collection_handler = webservice.CollectionHandler(
            r'/v1/data-sources',
            data_sources)
        resource_mgr.register_handler(ds_collection_handler)

        # Setup /v1/data-sources/<ds_id>
        ds_path = r'/v1/data-sources/(?P<ds_id>[^/]+)'
        ds_element_handler = webservice.ElementHandler(ds_path, data_sources)
        resource_mgr.register_handler(ds_element_handler)

        # Setup /v1/data-sources/<ds_id>/schema
        schema = process_dict['api-schema']
        schema_path = "%s/schema" % ds_path
        schema_element_handler = webservice.ElementHandler(schema_path, schema)
        resource_mgr.register_handler(schema_element_handler)

        # Setup /v1/data-sources/<ds_id>/tables/<table_id>/spec
        table_schema_path = "%s/tables/(?P<table_id>[^/]+)/spec" % ds_path
        table_schema_element_handler = webservice.ElementHandler(
            table_schema_path,
            schema)
        resource_mgr.register_handler(table_schema_element_handler)

        # Setup action handlers
        actions = process_dict['api-action']
        ds_actions_path = "%s/actions" % ds_path
        ds_actions_collection_handler = webservice.CollectionHandler(
            ds_actions_path, actions)
        resource_mgr.register_handler(ds_actions_collection_handler)

        # Setup status handlers
        statuses = process_dict['api-status']
        ds_status_path = "%s/status" % ds_path
        ds_status_element_handler = webservice.ElementHandler(ds_status_path,
                                                              statuses)
        resource_mgr.register_handler(ds_status_element_handler)
        policy_status_path = "%s/status" % policy_path
        policy_status_element_handler = webservice.ElementHandler(
            policy_status_path,
            statuses)
        resource_mgr.register_handler(policy_status_element_handler)
        rule_status_path = "%s/status" % rule_path
        rule_status_element_handler = webservice.ElementHandler(
            rule_status_path,
            statuses)
        resource_mgr.register_handler(rule_status_element_handler)

        tables = process_dict['api-table']
        tables_path = "(%s|%s)/tables" % (ds_path, policy_path)
        table_collection_handler = webservice.CollectionHandler(
            tables_path,
            tables)
        resource_mgr.register_handler(table_collection_handler)
        table_path = "%s/(?P<table_id>[^/]+)" % tables_path
        table_element_handler = webservice.ElementHandler(table_path, tables)
        resource_mgr.register_handler(table_element_handler)

        table_rows = process_dict['api-row']
        rows_path = "%s/rows" % table_path
        row_collection_handler = webservice.CollectionHandler(
            rows_path,
            table_rows, allow_update=True)
        resource_mgr.register_handler(row_collection_handler)
        row_path = "%s/(?P<row_id>[^/]+)" % rows_path
        row_element_handler = webservice.ElementHandler(row_path, table_rows)
        resource_mgr.register_handler(row_element_handler)

        # Setup /v1/system/datasource-drivers
        system = process_dict['api-system']
        # NOTE(arosen): start url out with datasource-drivers since we don't
        # yet implement /v1/system/ yet.
        system_collection_handler = webservice.CollectionHandler(
            r'/v1/system/drivers',
            system)
        resource_mgr.register_handler(system_collection_handler)

        # Setup /v1/system/datasource-drivers/<driver_id>
        driver_path = r'/v1/system/drivers/(?P<driver_id>[^/]+)'
        driver_element_handler = webservice.ElementHandler(driver_path, system)
        resource_mgr.register_handler(driver_element_handler)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from tempest import config
from tempest.lib import exceptions
from tempest import test

from congress_tempest_tests.tests.scenario import helper
from congress_tempest_tests.tests.scenario import manager_congress

CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestNovaDriver(manager_congress.ScenarioPolicyBase):

    @classmethod
    def skip_checks(cls):
        super(TestNovaDriver, cls).skip_checks()
        if not CONF.service_available.nova:
            skip_msg = ("%s skipped as nova is not available" % cls.__name__)
            raise cls.skipException(skip_msg)

        if not (CONF.network.tenant_networks_reachable
                or CONF.network.public_network_id):
            msg = ('Either tenant_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            cls.enabled = False
            raise cls.skipException(msg)

    def setUp(self):
        super(TestNovaDriver, self).setUp()
        self.keypairs = {}
        self.servers = []
        self.datasource_id = manager_congress.get_datasource_id(
            self.admin_manager.congress_client, 'nova')

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_nova_datasource_driver_servers(self):
        self._setup_network_and_servers()

        server_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'servers')['columns'])
        # Convert some of the column names.

        def convert_col(col):
            if col == 'host_id':
                return 'hostId'
            elif col == 'image_id':
                return 'image'
            elif col == 'flavor_id':
                return 'flavor'
            elif col == 'zone':
                return 'OS-EXT-AZ:availability_zone'
            elif col == 'host_name':
                return 'OS-EXT-SRV-ATTR:hypervisor_hostname'
            else:
                return col

        keys = [convert_col(c['name']) for c in server_schema]

        @helper.retry_on_exception
        def _check_data_table_nova_servers():
            results = (
                self.admin_manager.congress_client.list_datasource_rows(
                    self.datasource_id, 'servers'))
            for row in results['results']:
                match = True
                for index in range(len(keys)):
                    if keys[index] in ['image', 'flavor']:
                        val = self.servers[0][keys[index]]['id']
                    # Test servers created doesn't have this attribute,
                    # so ignoring the same in tempest tests.
                    elif keys[index] in \
                            ['OS-EXT-SRV-ATTR:hypervisor_hostname']:
                        continue
                    else:
                        val = self.servers[0][keys[index]]

                    if row['data'][index] != val:
                        match = False
                        break
                if match:
                    return True
            return False

        if not test.call_until_true(func=_check_data_table_nova_servers,
                                    duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_nova_datasource_driver_flavors(self):

        @helper.retry_on_exception
        def _check_data_table_nova_flavors():
            # Fetch data from nova each time, because this test may start
            # before nova has all the users.
            flavors = self.flavors_client.list_flavors(detail=True)
            flavor_id_map = {}
            for flavor in flavors['flavors']:
                flavor_id_map[flavor['id']] = flavor

            results = (
                self.admin_manager.congress_client.list_datasource_rows(
                    self.datasource_id, 'flavors'))
            # TODO(alexsyip): Not sure what the following OS-FLV-EXT-DATA:
            # prefix is for.
            keys = ['id', 'name', 'vcpus', 'ram', 'disk',
                    'OS-FLV-EXT-DATA:ephemeral', 'rxtx_factor']
            for row in results['results']:
                match = True
                try:
                    flavor_row = flavor_id_map[row['data'][0]]
                except KeyError:
                    return False
                for index in range(len(keys)):
                    if row['data'][index] != flavor_row[keys[index]]:
                        match = False
                        break
                if match:
                    return True
            return False

        if not test.call_until_true(func=_check_data_table_nova_flavors,
                                    duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")



# Copyright (c) 2014 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

import testtools

import congress.utils as utils


class UtilsTest(testtools.TestCase):

    def test_value_to_congress(self):
        self.assertEqual("abc", utils.value_to_congress("abc"))
        self.assertEqual("True", utils.value_to_congress(True))
        self.assertEqual("False", utils.value_to_congress(False))
        self.assertEqual(0, utils.value_to_congress(0))
        self.assertEqual(1, utils.value_to_congress(1))
        self.assertEqual(123, utils.value_to_congress(123))
        if sys.version < '3':
            self.assertEqual(456.0, utils.value_to_congress(456.0))



# Copyright 2014 Plexxi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import uuid


class d6msg(object):
    def __init__(self,
                 key="",
                 replyTo="",
                 correlationId="",
                 type="",
                 dataindex="",
                 body={},
                 srcmsg={}):

        self.header = {}

        self.body = body

        self.replyTo = replyTo
        self.type = type

        if srcmsg:
            self.key = srcmsg.replyTo
            self.correlationId = srcmsg.correlationId
            self.header['dataindex'] = srcmsg.header['dataindex']
        else:
            self.key = key
            self.header['dataindex'] = dataindex
            if correlationId:
                self.correlationId = correlationId
            else:
                newuuid = uuid.uuid4()
                self.correlationId = str(newuuid)

    def __str__(self):
        return ("<to:{}, from:{}, corrId:{}, type:{}, dataindex:{}, "
                "body:{}>").format(
                    self.key, self.replyTo, self.correlationId, self.type,
                    self.header['dataindex'], str(self.body))



# Copyright 2012 OpenStack Foundation
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from tempest import config
from tempest.lib import exceptions
from tempest import test

from congress_tempest_tests.tests.scenario import helper
from congress_tempest_tests.tests.scenario import manager_congress

import random
import string


CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestPolicyBasicOps(manager_congress.ScenarioPolicyBase):

    @classmethod
    def skip_checks(cls):
        super(TestPolicyBasicOps, cls).skip_checks()
        if not (CONF.network.tenant_networks_reachable
                or CONF.network.public_network_id):
            msg = ('Either tenant_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            cls.enabled = False
            raise cls.skipException(msg)

    def setUp(self):
        super(TestPolicyBasicOps, self).setUp()
        self.keypairs = {}
        self.servers = []

    def _create_random_policy(self):
        policy_name = "nova_%s" % ''.join(random.choice(string.lowercase)
                                          for x in range(10))
        body = {"name": policy_name}
        resp = self.admin_manager.congress_client.create_policy(body)
        self.addCleanup(self.admin_manager.congress_client.delete_policy,
                        resp['id'])
        return resp['name']

    def _create_policy_rule(self, policy_name, rule, rule_name=None,
                            comment=None):
        body = {'rule': rule}
        if rule_name:
            body['name'] = rule_name
        if comment:
            body['comment'] = comment
        client = self.admin_manager.congress_client
        response = client.create_policy_rule(policy_name, body)
        if response:
            self.addCleanup(client.delete_policy_rule, policy_name,
                            response['id'])
            return response
        else:
            raise Exception('Failed to create policy rule (%s, %s)'
                            % (policy_name, rule))

    def _create_test_server(self, name=None):
        image_ref = CONF.compute.image_ref
        flavor_ref = CONF.compute.flavor_ref
        keypair = self.create_keypair()
        security_group = self._create_security_group()
        security_groups = [{'name': security_group['name']}]
        create_kwargs = {'key_name': keypair['name'],
                         'security_groups': security_groups}
        instance = self.create_server(name=name,
                                      image_id=image_ref,
                                      flavor=flavor_ref,
                                      wait_until='ACTIVE',
                                      **create_kwargs)
        return instance

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_execution_action(self):
        metadata = {'testkey1': 'value3'}
        res = {'meta': {'testkey1': 'value3'}}
        server = self._create_test_server()
        congress_client = self.admin_manager.congress_client
        servers_client = self.admin_manager.servers_client
        policy = self._create_random_policy()
        service = 'nova'
        action = 'servers.set_meta'
        action_args = {'args': {'positional': [],
                                'named': {'server': server['id'],
                                          'metadata': metadata}}}
        body = action_args

        f = lambda: servers_client.show_server_metadata_item(server['id'],
                                                             'testkey1')
        # execute via datasource api
        body.update({'name': action})
        congress_client.execute_datasource_action(service, "execute", body)
        helper.retry_check_function_return_value(f, res)

        # execute via policy api
        body.update({'name': service + ':' + action})
        congress_client.execute_policy_action(policy, "execute", False,
                                              False, body)
        helper.retry_check_function_return_value(f, res)

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_policy_basic_op(self):
        self._setup_network_and_servers()
        body = {"rule": "port_security_group(id, security_group_name) "
                        ":-neutronv2:ports(id, tenant_id, name, network_id,"
                        "mac_address, admin_state_up, status, device_id, "
                        "device_owner),"
                        "neutronv2:security_group_port_bindings(id, "
                        "security_group_id), neutronv2:security_groups("
                        "security_group_id, tenant_id1, security_group_name,"
                        "description)"}
        results = self.admin_manager.congress_client.create_policy_rule(
            'classification', body)
        rule_id = results['id']
        self.addCleanup(
            self.admin_manager.congress_client.delete_policy_rule,
            'classification', rule_id)

        # Find the ports of on this server
        ports = self._list_ports(device_id=self.servers[0]['id'])

        def check_data():
            results = self.admin_manager.congress_client.list_policy_rows(
                'classification', 'port_security_group')
            for row in results['results']:
                if (row['data'][0] == ports[0]['id'] and
                    row['data'][1] ==
                        self.servers[0]['security_groups'][0]['name']):
                        return True
            else:
                return False

        if not test.call_until_true(func=check_data,
                                    duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_reactive_enforcement(self):
        servers_client = self.admin_manager.servers_client
        server_name = 'server_under_test'
        server = self._create_test_server(name=server_name)
        policy_name = self._create_random_policy()
        meta_key = 'meta_test_key1'
        meta_val = 'value1'
        meta_data = {'meta': {meta_key: meta_val}}
        rules = [
            'execute[nova:servers_set_meta(id, "%s", "%s")] :- '
            'test_servers(id)' % (meta_key, meta_val),
            'test_servers(id) :- '
            'nova:servers(id, name, host_id, status, tenant_id,'
            'user_id, image_id, flavor_id, zone, host_name),'
            'equal(name, "%s")' % server_name]

        for rule in rules:
            self._create_policy_rule(policy_name, rule)
        f = lambda: servers_client.show_server_metadata_item(server['id'],
                                                             meta_key)
        helper.retry_check_function_return_value(f, meta_data)


class TestCongressDataSources(manager_congress.ScenarioPolicyBase):

    @classmethod
    def skip_checks(cls):
        super(TestCongressDataSources, cls).skip_checks()
        if not (CONF.network.tenant_networks_reachable
                or CONF.network.public_network_id):
            msg = ('Either tenant_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            cls.enabled = False
            raise cls.skipException(msg)

    def test_all_loaded_datasources_are_initialized(self):
        datasources = self.admin_manager.congress_client.list_datasources()

        def _check_all_datasources_are_initialized():
            for datasource in datasources['results']:
                results = (
                    self.admin_manager.congress_client.list_datasource_status(
                        datasource['id']))
                if results['initialized'] != 'True':
                    return False
            return True

        if not test.call_until_true(
            func=_check_all_datasources_are_initialized,
                duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    def test_all_datasources_have_tables(self):
        datasources = self.admin_manager.congress_client.list_datasources()

        def check_data():
            for datasource in datasources['results']:
                results = (
                    self.admin_manager.congress_client.list_datasource_tables(
                        datasource['id']))
                # NOTE(arosen): if there are no results here we return false as
                # there is something wrong with a driver as it doesn't expose
                # any tables.
                if not results['results']:
                    return False
            return True

        if not test.call_until_true(func=check_data,
                                    duration=100, sleep_for=5):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")
